import inspect
import logging
import os
import random
import re
import tempfile
import zipfile
from abc import abstractmethod
from io import BytesIO
from typing import Dict, Generic, List, Optional, Sequence, Union

import torch
from torch.nn import Parameter, ParameterList
from transformers import (
    CONFIG_MAPPING,
    AutoConfig,
    AutoModel,
    AutoTokenizer,
    PretrainedConfig,
    PreTrainedTokenizer,
)
from transformers.tokenization_utils_base import LARGE_INTEGER

import flair
from flair.data import DT, Sentence, Token

log = logging.getLogger("flair")


class Embeddings(torch.nn.Module, Generic[DT]):
    """Abstract base class for all embeddings. Every new type of embedding must implement these methods."""

    def __init__(self):
        """Set some attributes that would otherwise result in errors. Overwrite these in your embedding class."""
        if not hasattr(self, "name"):
            self.name: str = "unnamed_embedding"
        if not hasattr(self, "static_embeddings"):
            # if the embeddings for a sentence are the same in each epoch, set this to True for improved efficiency
            self.static_embeddings = False
        super().__init__()

    @property
    @abstractmethod
    def embedding_length(self) -> int:
        """Returns the length of the embedding vector."""
        raise NotImplementedError

    @property
    @abstractmethod
    def embedding_type(self) -> str:
        raise NotImplementedError

    def embed(self, data_points: Union[DT, List[DT]]) -> List[DT]:
        """Add embeddings to all words in a list of sentences. If embeddings are already added, updates only if embeddings
        are non-static."""

        # if only one sentence is passed, convert to list of sentence
        if not isinstance(data_points, list):
            data_points = [data_points]

        if not self._everything_embedded(data_points) or not self.static_embeddings:
            self._add_embeddings_internal(data_points)

        return data_points

    def _everything_embedded(self, data_points: Sequence[DT]) -> bool:
        for data_point in data_points:
            if self.name not in data_point._embeddings.keys():
                return False
        return True

    @abstractmethod
    def _add_embeddings_internal(self, sentences: List[DT]):
        """Private method for adding embeddings to all words in a list of sentences."""
        pass

    def get_names(self) -> List[str]:
        """Returns a list of embedding names. In most cases, it is just a list with one item, namely the name of
        this embedding. But in some cases, the embedding is made up by different embeddings (StackedEmbedding).
        Then, the list contains the names of all embeddings in the stack."""
        return [self.name]

    def get_named_embeddings_dict(self) -> Dict:
        return {self.name: self}

    @staticmethod
    def get_instance_parameters(locals: dict) -> dict:
        class_definition = locals.get("__class__")
        instance_parameter_names = set(inspect.signature(class_definition.__init__).parameters)  # type: ignore
        instance_parameter_names.remove("self")
        instance_parameter_names.add("__class__")
        instance_parameters = {
            class_attribute: attribute_value
            for class_attribute, attribute_value in locals.items()
            if class_attribute in instance_parameter_names
        }
        return instance_parameters


class ScalarMix(torch.nn.Module):
    """
    Computes a parameterised scalar mixture of N tensors.
    This method was proposed by Liu et al. (2019) in the paper:
    "Linguistic Knowledge and Transferability of Contextual Representations" (https://arxiv.org/abs/1903.08855)

    The implementation is copied and slightly modified from the allennlp repository and is licensed under Apache 2.0.
    It can be found under:
    https://github.com/allenai/allennlp/blob/master/allennlp/modules/scalar_mix.py.
    """

    def __init__(self, mixture_size: int, trainable: bool = False) -> None:
        """
        Inits scalar mix implementation.
        ``mixture = gamma * sum(s_k * tensor_k)`` where ``s = softmax(w)``, with ``w`` and ``gamma`` scalar parameters.
        :param mixture_size: size of mixtures (usually the number of layers)
        """
        super(ScalarMix, self).__init__()
        self.mixture_size = mixture_size

        initial_scalar_parameters = [0.0] * mixture_size

        self.scalar_parameters = ParameterList(
            [
                Parameter(
                    torch.tensor(
                        [initial_scalar_parameters[i]],
                        dtype=torch.float,
                        device=flair.device,
                    ),
                    requires_grad=trainable,
                )
                for i in range(mixture_size)
            ]
        )
        self.gamma = Parameter(
            torch.tensor(
                [1.0],
                dtype=torch.float,
                device=flair.device,
            ),
            requires_grad=trainable,
        )

    def forward(self, tensors: List[torch.Tensor]) -> torch.Tensor:
        """
        Computes a weighted average of the ``tensors``.  The input tensors an be any shape
        with at least two dimensions, but must all be the same shape.
        :param tensors: list of input tensors
        :return: computed weighted average of input tensors
        """
        if len(tensors) != self.mixture_size:
            log.error(
                "{} tensors were passed, but the module was initialized to mix {} tensors.".format(
                    len(tensors), self.mixture_size
                )
            )

        normed_weights = torch.nn.functional.softmax(
            torch.cat([parameter for parameter in self.scalar_parameters]), dim=0
        )
        normed_weights = torch.split(normed_weights, split_size_or_sections=1)

        pieces = []
        for weight, tensor in zip(normed_weights, tensors):
            pieces.append(weight * tensor)
        return self.gamma * sum(pieces)


class TransformerEmbedding(Embeddings[Sentence]):
    def __init__(
        self,
        model: str = "bert-base-uncased",
        fine_tune: bool = True,
        layers: str = "-1",
        layer_mean: bool = True,
        subtoken_pooling: str = "first",
        cls_pooling: str = "cls",
        is_token_embedding: bool = True,
        is_document_embedding: bool = True,
        allow_long_sentences: bool = False,
        use_context: Union[bool, int] = False,
        respect_document_boundaries: bool = True,
        context_dropout: float = 0.5,
        saved_config: Optional[PretrainedConfig] = None,
        tokenizer_data: Optional[BytesIO] = None,
        name: Optional[str] = None,
        **kwargs,
    ):
        self.instance_parameters = self.get_instance_parameters(locals=locals())
        del self.instance_parameters["saved_config"]
        del self.instance_parameters["tokenizer_data"]
        super().__init__()
        # temporary fix to disable tokenizer parallelism warning
        # (see https://stackoverflow.com/questions/62691279/how-to-disable-tokenizers-parallelism-true-false-warning)
        os.environ["TOKENIZERS_PARALLELISM"] = "false"

        # do not print transformer warnings as these are confusing in this case
        from transformers import logging

        logging.set_verbosity_error()

        if tokenizer_data is None:
            # load tokenizer and transformer model
            self.tokenizer: PreTrainedTokenizer = AutoTokenizer.from_pretrained(model, **kwargs)
        else:
            # load tokenizer from inmemory zip-file
            self.tokenizer = self._tokenizer_from_bytes(tokenizer_data)

        if saved_config is None:
            config = AutoConfig.from_pretrained(model, output_hidden_states=True, **kwargs)
            self.model = AutoModel.from_pretrained(model, config=config)
        else:
            self.model = AutoModel.from_config(saved_config, **kwargs)

        self.truncate = True

        if self.tokenizer.model_max_length > LARGE_INTEGER:
            allow_long_sentences = False
            self.truncate = False

        self.stride = self.tokenizer.model_max_length // 2 if allow_long_sentences else 0
        self.allow_long_sentences = allow_long_sentences
        self.use_lang_emb = hasattr(self.model, "use_lang_emb") and self.model.use_lang_emb

        # model name
        if name is None:
            self.name = "transformer-" + str(model)
        else:
            self.name = name
        self.base_model_name = str(model)

        self.token_embedding = is_token_embedding
        self.document_embedding = is_document_embedding

        if not self.token_embedding and not self.document_embedding:
            raise ValueError("either 'is_token_embedding' or 'is_document_embedding' needs to be set.")

        if self.document_embedding and cls_pooling not in ["cls", "max", "mean"]:
            raise ValueError(f"Document Pooling operation `{cls_pooling}` is not defined for TransformerEmbedding")

        if self.token_embedding and subtoken_pooling not in ["first", "last", "first_last", "mean"]:
            raise ValueError(f"Subtoken Pooling operation `{subtoken_pooling}` is not defined for TransformerEmbedding")

        if self.document_embedding and cls_pooling == "cls" and allow_long_sentences:
            log.warning(
                "Using long sentences for Document embeddings is only beneficial for cls_pooling types 'mean' and 'max "
            )

        if isinstance(use_context, bool):
            self.context_length: int = 64 if use_context else 0
        else:
            self.context_length = use_context

        self.context_dropout = context_dropout
        self.respect_document_boundaries = respect_document_boundaries

        self.to(flair.device)

        # embedding parameters
        if layers == "all":
            # send mini-token through to check how many layers the model has
            hidden_states = self.model(torch.tensor([1], device=flair.device).unsqueeze(0))[-1]
            self.layer_indexes = list(range(len(hidden_states)))
        else:
            self.layer_indexes = list(map(int, layers.split(",")))

        self.cls_pooling = cls_pooling
        self.subtoken_pooling = subtoken_pooling
        self.layer_mean = layer_mean
        self.fine_tune = fine_tune
        self.static_embeddings = not self.fine_tune

        # return length
        self.embedding_length_internal = self._calculate_embedding_length()

        self.special_tokens = []
        # check if special tokens exist to circumvent error message
        if self.tokenizer._bos_token:
            self.special_tokens.append(self.tokenizer.bos_token)
        if self.tokenizer._cls_token:
            self.special_tokens.append(self.tokenizer.cls_token)

        # most models have an initial BOS token, except for XLNet, T5 and GPT2
        self.begin_offset = self._get_begin_offset_of_tokenizer()
        self.initial_cls_token: bool = self._has_initial_cls_token()

        # when initializing, embeddings are in eval mode by default
        self.eval()

    @property
    def embedding_length(self) -> int:
        if not hasattr(self, "embedding_length_internal"):
            self.embedding_length_internal = self._calculate_embedding_length()

        return self.embedding_length_internal

    def _has_initial_cls_token(self) -> bool:
        # most models have CLS token as last token (GPT-1, GPT-2, TransfoXL, XLNet, XLM), but BERT is initial
        tokens = self.tokenizer.encode("a")
        return tokens[0] == self.tokenizer.cls_token_id

    def _get_begin_offset_of_tokenizer(self) -> int:
        test_string = "a"
        tokens_with_special = self.tokenizer.encode(test_string)
        tokens_without_special = self.tokenizer.encode(test_string, add_special_tokens=False)
        normal_count = len(tokens_without_special)

        for begin_offset in range(len(tokens_with_special) - normal_count):
            if tokens_with_special[begin_offset : begin_offset + normal_count] == tokens_without_special:
                return begin_offset
        log.warning(
            f"Could not determine the begin offset of the tokenizer for transformer model {self.name}, assuming 0"
        )
        return 0

    def _calculate_embedding_length(self) -> int:
        if not self.layer_mean:
            length = len(self.layer_indexes) * self.model.config.hidden_size
        else:
            length = self.model.config.hidden_size

        # in case of doubt: token embedding has higher priority than document embedding
        if self.token_embedding and self.subtoken_pooling == "first_last":
            length *= 2
            if self.document_embedding:
                log.warning(
                    "Token embedding length and Document embedding length vary, due to `first_last` subtoken pooling, this might not be supported"
                )
        return length

    @property
    def embedding_type(self) -> str:
        # in case of doubt: token embedding has higher priority than document embedding
        return "word-level" if self.token_embedding else "sentence-level"

    def _tokenizer_from_bytes(self, zip_data: BytesIO) -> PreTrainedTokenizer:
        zip_obj = zipfile.ZipFile(zip_data)
        with tempfile.TemporaryDirectory() as temp_dir:
            zip_obj.extractall(temp_dir)
            return AutoTokenizer.from_pretrained(temp_dir)

    def _tokenizer_bytes(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            files = list(self.tokenizer.save_pretrained(temp_dir))
            if self.tokenizer.is_fast:
                vocab_files = self.tokenizer.slow_tokenizer_class.vocab_files_names.values()
                files = [f for f in files if all(v not in f for v in vocab_files)]
            zip_data = BytesIO()
            zip_obj = zipfile.ZipFile(zip_data, "w")
            for f in files:
                # transformers returns the "added_tokens.json" even if it doesn't create it
                if os.path.exists(f):
                    zip_obj.write(f, os.path.relpath(f, temp_dir))

        zip_data.seek(0)
        return zip_data

    @staticmethod
    def _remove_special_markup(text: str):
        # remove special markup
        text = re.sub("^Ġ", "", text)  # RoBERTa models
        text = re.sub("^##", "", text)  # BERT models
        text = re.sub("^▁", "", text)  # XLNet models
        text = re.sub("</w>$", "", text)  # XLM models
        return text

    def _get_processed_token_text(self, token: Token) -> str:
        pieces = self.tokenizer.tokenize(token.text)
        token_text = ""
        for piece in pieces:
            token_text += self._remove_special_markup(piece)
        token_text = token_text.lower()
        return token_text

    def __getstate__(self):
        config_dict = self.model.config.to_dict()

        tokenizer_data = self._tokenizer_bytes()

        model_state = {
            "model": self.base_model_name,
            "fine_tune": self.fine_tune,
            "layers": ",".join(map(str, self.layer_indexes)),
            "layer_mean": self.layer_mean,
            "subtoken_pooling": self.subtoken_pooling,
            "cls_pooling": self.cls_pooling,
            "is_token_embedding": self.token_embedding,
            "is_document_embedding": self.document_embedding,
            "allow_long_sentences": self.allow_long_sentences,
            "config_state_dict": config_dict,
            "tokenizer_data": tokenizer_data,
            "name": self.name,
            "context_length": self.context_length,
            "respect_document_boundaries": self.respect_document_boundaries,
            "context_dropout": self.context_dropout,
        }

        return model_state

    def __setstate__(self, state):
        config_state_dict = state.pop("config_state_dict", None)
        model_state_dict = state.pop("model_state_dict", None)

        # legacy TransformerDocumentEmbedding
        state.pop("batch_size", None)
        state.pop("embedding_length_internal", None)
        # legacy TransformerTokenEmbedding
        state.pop("memory_effective_training", None)

        if "base_model_name" in state:
            state["model"] = state.pop("base_model_name")

        state["use_context"] = state.pop("context_length", False)

        if "layer_indexes" in state:
            layer_indexes = state.pop("layer_indexes")
            state["layers"] = ",".join(map(str, layer_indexes))

        if "use_scalar_mix" in state:
            # legacy Flair <= 0.7
            state["layer_mean"] = state.pop("use_scalar_mix")

        if "is_token_embedding" not in state:
            # legacy TransformerTokenEmbedding
            state["is_token_embedding"] = "pooling_operation" in state

        if "is_document_embedding" not in state:
            # Legacy TransformerDocumentEmbedding
            state["is_document_embedding"] = "pooling" in state

        if "pooling_operation" in state:
            # legacy TransformerTokenEmbedding
            state["subtoken_pooling"] = state.pop("pooling_operation")

        if "cls_operation" in state:
            # legacy TransformerDocumentEmbedding
            state["cls_pooling"] = state.pop("pooling")

        config = None

        if config_state_dict:
            model_type = config_state_dict.get("model_type", "bert")
            config_class = CONFIG_MAPPING[model_type]
            config = config_class.from_dict(config_state_dict)

        embedding = self.create_from_state(saved_config=config, **state)

        # copy values from new embedding
        for key in embedding.__dict__.keys():
            self.__dict__[key] = embedding.__dict__[key]

        if model_state_dict:
            self.model.load_state_dict(model_state_dict)

    @classmethod
    def create_from_state(cls, **state):
        return cls(**state)

    def _reconstruct_tokens_from_subtokens(self, sentence, subtokens):
        word_iterator = iter(sentence)
        token = next(word_iterator)
        token_text = self._get_processed_token_text(token)
        token_subtoken_lengths = []
        reconstructed_token = ""
        subtoken_count = 0
        # iterate over subtokens and reconstruct tokens
        for subtoken_id, subtoken in enumerate(subtokens):

            # remove special markup
            subtoken = self._remove_special_markup(subtoken)

            # TODO check if this is necessary is this method is called before prepare_for_model
            # check if reconstructed token is special begin token ([CLS] or similar)
            if subtoken in self.special_tokens and subtoken_id == 0:
                continue

            # some BERT tokenizers somehow omit words - in such cases skip to next token
            if subtoken_count == 0 and not token_text.startswith(subtoken.lower()):

                while True:
                    token_subtoken_lengths.append(0)
                    token = next(word_iterator)
                    token_text = self._get_processed_token_text(token)
                    if token_text.startswith(subtoken.lower()):
                        break

            subtoken_count += 1

            # append subtoken to reconstruct token
            reconstructed_token = reconstructed_token + subtoken

            # check if reconstructed token is the same as current token
            if reconstructed_token.lower() == token_text:

                # if so, add subtoken count
                token_subtoken_lengths.append(subtoken_count)

                # reset subtoken count and reconstructed token
                reconstructed_token = ""
                subtoken_count = 0

                # break from loop if all tokens are accounted for
                if len(token_subtoken_lengths) < len(sentence):
                    token = next(word_iterator)
                    token_text = self._get_processed_token_text(token)
                else:
                    break

        # if tokens are unaccounted for
        while len(token_subtoken_lengths) < len(sentence) and len(token.text) == 1:
            token_subtoken_lengths.append(0)
            if len(token_subtoken_lengths) == len(sentence):
                break
            token = next(word_iterator)

        # check if all tokens were matched to subtokens
        if token != sentence[-1]:
            log.error(f"Tokenization MISMATCH in sentence '{sentence.to_tokenized_string()}'")
            log.error(f"Last matched: '{token}'")
            log.error(f"Last sentence: '{sentence[-1]}'")
            log.error(f"subtokenized: '{subtokens}'")
        return token_subtoken_lengths

    def _gather_tokenized_strings(self, sentences: List[Sentence]):
        tokenized_sentences = []
        all_token_subtoken_lengths = []
        subtoken_lengths = []
        for sentence in sentences:

            # subtokenize the sentence
            tokenized_string = sentence.to_tokenized_string()

            # transformer specific tokenization
            subtokenized_sentence = self.tokenizer.tokenize(tokenized_string)

            # set zero embeddings for empty sentences and exclude
            if len(subtokenized_sentence) == 0:
                if self.token_embedding:
                    for token in sentence:
                        token.set_embedding(self.name, torch.zeros(self.embedding_length))
                if self.document_embedding:
                    sentence.set_embedding(self.name, torch.zeros(self.embedding_length))
                continue

            if self.token_embedding:
                # determine into how many subtokens each token is split
                all_token_subtoken_lengths.append(
                    self._reconstruct_tokens_from_subtokens(sentence, subtokenized_sentence)
                )
            subtoken_lengths.append(len(subtokenized_sentence))

            # remember tokenized sentences and their subtokenization
            tokenized_sentences.append(tokenized_string)
        return tokenized_sentences, all_token_subtoken_lengths, subtoken_lengths

    def _build_transformer_model_inputs(self, batch_encoding, tokenized_sentences, sentences):
        model_kwargs = {}
        input_ids = batch_encoding["input_ids"].to(flair.device)

        # Models such as FNet do not have an attention_mask
        if "attention_mask" in batch_encoding:
            model_kwargs["attention_mask"] = batch_encoding["attention_mask"].to(flair.device)

        # set language IDs for XLM-style transformers
        if self.use_lang_emb and self.tokenizer.lang2id is not None:
            model_kwargs["langs"] = torch.zeros_like(input_ids, dtype=input_ids.dtype)
            if not self.allow_long_sentences:
                for s_id, sentence_text in enumerate(tokenized_sentences):
                    lang_id = self.tokenizer.lang2id.get(sentences[s_id].get_language_code(), 0)
                    model_kwargs["langs"][s_id] = lang_id
            else:
                sentence_part_lengths = torch.unique(
                    batch_encoding["overflow_to_sample_mapping"],
                    return_counts=True,
                    sorted=True,
                )[1].tolist()
                sentence_idx = 0
                for sentence, part_length in zip(sentences, sentence_part_lengths):
                    lang_id = self.tokenizer.lang2id.get(sentence.get_language_code(), 0)
                    model_kwargs["langs"][sentence_idx : sentence_idx + part_length] = lang_id
                    sentence_idx += part_length

        return input_ids, model_kwargs

    def _combine_strided_sentences(
        self, hidden_states: torch.Tensor, sentence_parts_lengths: torch.Tensor
    ) -> List[torch.Tensor]:
        sentence_idx_offset = 0
        sentence_hidden_states = []
        for nr_sentence_parts in sentence_parts_lengths:
            sentence_hidden_state = hidden_states[:, sentence_idx_offset, ...]
            sentence_idx_offset += 1

            for i in range(1, nr_sentence_parts):
                remainder_sentence_hidden_state = hidden_states[:, sentence_idx_offset, ...]
                sentence_idx_offset += 1
                sentence_hidden_state = torch.cat(
                    (
                        sentence_hidden_state[:, : -1 - self.stride // 2, :],
                        remainder_sentence_hidden_state[:, 1 + self.stride // 2 :, :],
                    ),
                    1,
                )
            sentence_hidden_states.append(sentence_hidden_state)
        return sentence_hidden_states

    def _try_document_embedding_shortcut(self, hidden_states, sentences):
        # cls first pooling can be done without recreating sentence hidden states
        if (
            self.document_embedding
            and not self.token_embedding
            and self.cls_pooling == "cls"
            and self.initial_cls_token
        ):
            embeddings_all_document_layers = hidden_states[:, :, 0]
            if self.layer_mean:
                document_embs = torch.mean(embeddings_all_document_layers, dim=0)
            else:
                document_embs = torch.flatten(embeddings_all_document_layers.permute((1, 0, 2)), 1)
            for (document_emb, sentence) in zip(document_embs, sentences):
                sentence.set_embedding(self.name, document_emb)
            return True
        return False

    def _extract_document_embeddings(self, sentence_hidden_states, sentences):
        for sentence_hidden_state, sentence in zip(sentence_hidden_states, sentences):
            if self.cls_pooling == "cls":
                index_of_cls_token = 0 if self.initial_cls_token else -1
                embedding_all_document_layers = sentence_hidden_state[:, index_of_cls_token, :]
            elif self.cls_pooling == "mean":
                embedding_all_document_layers = sentence_hidden_state.mean(dim=1)
            elif self.cls_pooling == "max":
                embedding_all_document_layers, _ = sentence_hidden_state.max(dim=1)
            else:
                raise ValueError(f"cls pooling method: `{self.cls_pooling}` is not implemented")
            if self.layer_mean:
                document_emb = torch.mean(embedding_all_document_layers, dim=0)
            else:
                document_emb = embedding_all_document_layers.view(-1)
            sentence.set_embedding(self.name, document_emb)

    def _extract_token_embeddings(self, sentence_hidden_states, sentences, all_token_subtoken_lengths):
        for sentence_hidden_state, sentence, subtoken_lengths in zip(
            sentence_hidden_states, sentences, all_token_subtoken_lengths
        ):
            subword_start_idx = self.begin_offset

            for token, n_subtokens in zip(sentence, subtoken_lengths):
                if n_subtokens == 0:
                    token.set_embedding(self.name, torch.zeros(self.embedding_length))
                    continue
                subword_end_idx = subword_start_idx + n_subtokens
                assert subword_start_idx < subword_end_idx <= sentence_hidden_state.size()[1]
                current_embeddings = sentence_hidden_state[:, subword_start_idx:subword_end_idx]
                subword_start_idx = subword_end_idx
                if self.subtoken_pooling == "first":
                    final_embedding = current_embeddings[:, 0]
                elif self.subtoken_pooling == "last":
                    final_embedding = current_embeddings[:, -1]
                elif self.subtoken_pooling == "first_last":
                    final_embedding = torch.cat([current_embeddings[:, 0], current_embeddings[:, -1]], dim=1)
                elif self.subtoken_pooling == "mean":
                    final_embedding = current_embeddings.mean(dim=1)
                else:
                    raise ValueError(f"subtoken pooling method: `{self.subtoken_pooling}` is not implemented")

                if self.layer_mean:
                    final_embedding = final_embedding.mean(dim=0)
                else:
                    final_embedding = torch.flatten(final_embedding)

                token.set_embedding(self.name, final_embedding)

    def _add_embeddings_to_sentences(self, sentences: List[Sentence]):
        tokenized_sentences, all_token_subtoken_lengths, subtoken_lengths = self._gather_tokenized_strings(sentences)

        # encode inputs
        batch_encoding = self.tokenizer(
            tokenized_sentences,
            stride=self.stride,
            return_overflowing_tokens=self.allow_long_sentences,
            truncation=self.truncate,
            padding=True,
            return_tensors="pt",
        )

        input_ids, model_kwargs = self._build_transformer_model_inputs(batch_encoding, tokenized_sentences, sentences)

        gradient_context = torch.enable_grad() if (self.fine_tune and self.training) else torch.no_grad()

        with gradient_context:
            hidden_states = self.model(input_ids, **model_kwargs)[-1]

            # make the tuple a tensor; makes working with it easier.
            hidden_states = torch.stack(hidden_states)

            # only use layers that will be outputted
            hidden_states = hidden_states[self.layer_indexes, :, :]

            if self._try_document_embedding_shortcut(hidden_states, sentences):
                return

            if self.allow_long_sentences:
                sentence_hidden_states = self._combine_strided_sentences(
                    hidden_states,
                    sentence_parts_lengths=torch.unique(
                        batch_encoding["overflow_to_sample_mapping"],
                        return_counts=True,
                        sorted=True,
                    )[1].tolist(),
                )
            else:
                sentence_hidden_states = list(hidden_states.permute((1, 0, 2, 3)))

            # remove padding tokens
            sentence_hidden_states = [
                sentence_hidden_state[:, : subtoken_length + 1, :]
                for (subtoken_length, sentence_hidden_state) in zip(subtoken_lengths, sentence_hidden_states)
            ]

            if self.document_embedding:
                self._extract_document_embeddings(sentence_hidden_states, sentences)

            if self.token_embedding:
                self._extract_token_embeddings(sentence_hidden_states, sentences, all_token_subtoken_lengths)

    def _expand_sentence_with_context(self, sentence):
        expand_context = not self.training or random.randint(1, 100) > (self.context_dropout * 100)

        left_context = []
        right_context = []

        if expand_context:
            left_context = sentence.left_context(self.context_length, self.respect_document_boundaries)
            right_context = sentence.right_context(self.context_length, self.respect_document_boundaries)

        expanded_sentence = Sentence(left_context + [t.text for t in sentence.tokens] + right_context)

        context_length = len(left_context)
        return expanded_sentence, context_length

    def _add_embeddings_internal(self, sentences: List[Sentence]):
        expanded_sentences = []
        context_offsets = []

        if self.context_length > 0:
            # set context if not set already
            previous_sentence = None
            for sentence in sentences:
                if sentence.is_context_set():
                    continue
                sentence._previous_sentence = previous_sentence
                sentence._next_sentence = None
                if previous_sentence:
                    previous_sentence._next_sentence = sentence
                previous_sentence = sentence

            for sentence in sentences:
                # create expanded sentence and remember context offsets
                expanded_sentence, context_offset = self._expand_sentence_with_context(sentence)
                expanded_sentences.append(expanded_sentence)
                context_offsets.append(context_offset)
        else:
            expanded_sentences.extend(sentences)

        self._add_embeddings_to_sentences(expanded_sentences)

        # move embeddings from context back to original sentence (if using context)
        if self.context_length > 0:
            for original_sentence, expanded_sentence, context_offset in zip(
                sentences, expanded_sentences, context_offsets
            ):
                if self.token_embedding:
                    for token_idx, token in enumerate(original_sentence):
                        token.set_embedding(
                            self.name,
                            expanded_sentence[token_idx + context_offset].get_embedding(self.name),
                        )
                if self.document_embedding:
                    original_sentence.set_embedding(self.name, expanded_sentence.get_embedding(self.name))



import shutil

import pytest
from hyperopt import hp
from torch.optim import SGD

import flair.datasets
from flair.embeddings import StackedEmbeddings, WordEmbeddings
from flair.hyperparameter import (
    Parameter,
    SearchSpace,
    SequenceTaggerParamSelector,
    TextClassifierParamSelector,
)

glove_embedding: WordEmbeddings = WordEmbeddings("glove")


@pytest.mark.integration
def test_sequence_tagger_param_selector(results_base_path, tasks_base_path):
    corpus = flair.datasets.ColumnCorpus(data_folder=tasks_base_path / "fashion", column_format={0: "text", 3: "ner"})

    # define search space
    search_space = SearchSpace()

    # sequence tagger parameter
    search_space.add(
        Parameter.EMBEDDINGS,
        hp.choice,
        options=[StackedEmbeddings([glove_embedding])],
    )
    search_space.add(Parameter.USE_CRF, hp.choice, options=[True, False])
    search_space.add(Parameter.DROPOUT, hp.uniform, low=0.25, high=0.75)
    search_space.add(Parameter.WORD_DROPOUT, hp.uniform, low=0.0, high=0.25)
    search_space.add(Parameter.LOCKED_DROPOUT, hp.uniform, low=0.0, high=0.5)
    search_space.add(Parameter.HIDDEN_SIZE, hp.choice, options=[64, 128])
    search_space.add(Parameter.RNN_LAYERS, hp.choice, options=[1, 2])

    # model trainer parameter
    search_space.add(Parameter.OPTIMIZER, hp.choice, options=[SGD])

    # training parameter
    search_space.add(Parameter.MINI_BATCH_SIZE, hp.choice, options=[4, 8, 32])
    search_space.add(Parameter.LEARNING_RATE, hp.uniform, low=0.01, high=1)
    search_space.add(Parameter.ANNEAL_FACTOR, hp.uniform, low=0.3, high=0.75)
    search_space.add(Parameter.PATIENCE, hp.choice, options=[3, 5])
    search_space.add(Parameter.WEIGHT_DECAY, hp.uniform, low=0.01, high=1)

    # find best parameter settings
    optimizer = SequenceTaggerParamSelector(corpus, "ner", results_base_path, max_epochs=2)
    optimizer.optimize(search_space, max_evals=2)

    # clean up results directory
    shutil.rmtree(results_base_path)
    del optimizer, search_space


@pytest.mark.integration
def test_text_classifier_param_selector(results_base_path, tasks_base_path):
    corpus = flair.datasets.ClassificationCorpus(tasks_base_path / "imdb", label_type="sentiment")
    label_type = "sentiment"

    search_space = SearchSpace()

    # document embeddings parameter
    search_space.add(Parameter.TRANSFORMER_MODEL, hp.choice, options=["sshleifer/tiny-distilbert-base-cased"])
    search_space.add(Parameter.LAYERS, hp.choice, options=["-1", "-2"])

    # training parameter
    search_space.add(Parameter.LEARNING_RATE, hp.uniform, low=0, high=1)
    search_space.add(Parameter.MINI_BATCH_SIZE, hp.choice, options=[4, 8, 16, 32])
    search_space.add(Parameter.ANNEAL_FACTOR, hp.uniform, low=0, high=0.75)
    search_space.add(Parameter.PATIENCE, hp.choice, options=[3, 5])

    param_selector = TextClassifierParamSelector(corpus, label_type, False, results_base_path, max_epochs=2)
    param_selector.optimize(search_space, max_evals=2)

    # clean up results directory
    shutil.rmtree(results_base_path)
    del param_selector, search_space



# Expose base classses
from .base import Embeddings, ScalarMix

# Expose document embedding classes
from .document import (
    DocumentCNNEmbeddings,
    DocumentEmbeddings,
    DocumentLMEmbeddings,
    DocumentPoolEmbeddings,
    DocumentRNNEmbeddings,
    DocumentTFIDFEmbeddings,
    SentenceTransformerDocumentEmbeddings,
    TransformerDocumentEmbeddings,
)

# Expose image embedding classes
from .image import (
    ConvTransformNetworkImageEmbeddings,
    IdentityImageEmbeddings,
    ImageEmbeddings,
    NetworkImageEmbeddings,
    PrecomputedImageEmbeddings,
)

# Expose legacy embedding classes
from .legacy import (
    BertEmbeddings,
    CamembertEmbeddings,
    CharLMEmbeddings,
    DocumentLSTMEmbeddings,
    DocumentMeanEmbeddings,
    ELMoTransformerEmbeddings,
    OpenAIGPT2Embeddings,
    OpenAIGPTEmbeddings,
    RoBERTaEmbeddings,
    TransformerXLEmbeddings,
    XLMEmbeddings,
    XLMRobertaEmbeddings,
    XLNetEmbeddings,
)

# Expose token embedding classes
from .token import (
    BPEmbSerializable,
    BytePairEmbeddings,
    CharacterEmbeddings,
    ELMoEmbeddings,
    FastTextEmbeddings,
    FlairEmbeddings,
    HashEmbeddings,
    MuseCrosslingualEmbeddings,
    NILCEmbeddings,
    OneHotEmbeddings,
    PooledFlairEmbeddings,
    StackedEmbeddings,
    TokenEmbeddings,
    TransformerWordEmbeddings,
    WordEmbeddings,
)

__all__ = [
    "Embeddings",
    "ScalarMix",
    "DocumentCNNEmbeddings",
    "DocumentEmbeddings",
    "DocumentLMEmbeddings",
    "DocumentPoolEmbeddings",
    "DocumentRNNEmbeddings",
    "DocumentTFIDFEmbeddings",
    "SentenceTransformerDocumentEmbeddings",
    "TransformerDocumentEmbeddings",
    "ConvTransformNetworkImageEmbeddings",
    "IdentityImageEmbeddings",
    "ImageEmbeddings",
    "NetworkImageEmbeddings",
    "PrecomputedImageEmbeddings",
    "BertEmbeddings",
    "CamembertEmbeddings",
    "CharLMEmbeddings",
    "DocumentLSTMEmbeddings",
    "DocumentMeanEmbeddings",
    "ELMoTransformerEmbeddings",
    "OpenAIGPT2Embeddings",
    "OpenAIGPTEmbeddings",
    "RoBERTaEmbeddings",
    "TransformerXLEmbeddings",
    "XLMEmbeddings",
    "XLMRobertaEmbeddings",
    "XLNetEmbeddings",
    "BPEmbSerializable",
    "BytePairEmbeddings",
    "CharacterEmbeddings",
    "ELMoEmbeddings",
    "FastTextEmbeddings",
    "FlairEmbeddings",
    "HashEmbeddings",
    "MuseCrosslingualEmbeddings",
    "NILCEmbeddings",
    "OneHotEmbeddings",
    "PooledFlairEmbeddings",
    "StackedEmbeddings",
    "TokenEmbeddings",
    "TransformerWordEmbeddings",
    "WordEmbeddings",
]



import pytest
from torch.optim import SGD
from torch.optim.adam import Adam

import flair.datasets
from flair.data import MultiCorpus, Sentence
from flair.embeddings import FlairEmbeddings, WordEmbeddings
from flair.models import SequenceTagger
from flair.trainers import ModelTrainer

turian_embeddings = WordEmbeddings("turian")
flair_embeddings = FlairEmbeddings("news-forward-fast")


@pytest.mark.integration
def test_load_use_tagger():
    loaded_model: SequenceTagger = SequenceTagger.load("ner-fast")

    sentence = Sentence("I love Berlin")
    sentence_empty = Sentence("       ")

    loaded_model.predict(sentence)
    loaded_model.predict([sentence, sentence_empty])
    loaded_model.predict([sentence_empty])

    # check that "Berlin" is predicted as LOC-Span
    assert len(sentence.get_spans("ner")) == 1
    assert sentence.get_spans("ner")[0].text == "Berlin"
    assert sentence.get_spans("ner")[0].tag == "LOC"

    # check that "Berlin" is predicted as S-LOC-Token when force_token_predictions=True
    sentence = Sentence("I love Berlin")
    loaded_model.predict(sentence, force_token_predictions=True)
    assert sentence.get_token(3).text == "Berlin"
    assert sentence.get_token(3).tag == "S-LOC"

    del loaded_model

    loaded_model: SequenceTagger = SequenceTagger.load("pos")

    loaded_model.predict(sentence)
    loaded_model.predict([sentence, sentence_empty])
    loaded_model.predict([sentence_empty])

    del loaded_model


@pytest.mark.integration
def test_load_use_tagger_keep_embedding():
    loaded_model: SequenceTagger = SequenceTagger.load("ner-fast")

    sentence = Sentence("I love Berlin")
    loaded_model.predict(sentence)
    for token in sentence:
        assert len(token.embedding.cpu().numpy()) == 0

    loaded_model.predict(sentence, embedding_storage_mode="cpu")
    for token in sentence:
        assert len(token.embedding.cpu().numpy()) > 0

    del loaded_model


@pytest.mark.integration
def test_all_tag_proba_embedding():
    loaded_model: SequenceTagger = SequenceTagger.load("ner-fast")

    sentence = Sentence("I love Berlin")
    loaded_model.predict(sentence, return_probabilities_for_all_classes=True)
    for token in sentence:
        assert len(token.get_tags_proba_dist(loaded_model.tag_type)) == len(loaded_model.label_dictionary)
        score_sum = 0
        for label in token.get_tags_proba_dist(loaded_model.tag_type):
            assert label.data_point == token
            score_sum += label.score
        assert abs(score_sum - 1.0) < 1.0e-5


@pytest.mark.integration
def test_train_load_use_tagger(results_base_path, tasks_base_path):
    corpus = flair.datasets.ColumnCorpus(data_folder=tasks_base_path / "fashion", column_format={0: "text", 3: "ner"})
    tag_dictionary = corpus.make_label_dictionary("ner")

    tagger: SequenceTagger = SequenceTagger(
        hidden_size=64,
        embeddings=turian_embeddings,
        tag_dictionary=tag_dictionary,
        tag_type="ner",
        use_crf=False,
    )

    # initialize trainer
    trainer: ModelTrainer = ModelTrainer(tagger, corpus)

    trainer.train(
        results_base_path,
        learning_rate=0.1,
        mini_batch_size=2,
        max_epochs=2,
        shuffle=False,
    )

    del trainer, tagger, tag_dictionary, corpus
    loaded_model: SequenceTagger = SequenceTagger.load(results_base_path / "final-model.pt")

    sentence = Sentence("I love Berlin")
    sentence_empty = Sentence("       ")

    loaded_model.predict(sentence)
    loaded_model.predict([sentence, sentence_empty])
    loaded_model.predict([sentence_empty])

    # clean up results directory
    del loaded_model


@pytest.mark.integration
def test_train_load_use_tagger_empty_tags(results_base_path, tasks_base_path):
    corpus = flair.datasets.ColumnCorpus(data_folder=tasks_base_path / "fashion", column_format={0: "text", 2: "ner"})
    tag_dictionary = corpus.make_label_dictionary("ner")

    tagger: SequenceTagger = SequenceTagger(
        hidden_size=64,
        embeddings=turian_embeddings,
        tag_dictionary=tag_dictionary,
        tag_type="ner",
        use_crf=False,
    )

    # initialize trainer
    trainer: ModelTrainer = ModelTrainer(tagger, corpus)

    trainer.train(
        results_base_path,
        learning_rate=0.1,
        mini_batch_size=2,
        max_epochs=2,
        shuffle=False,
    )

    del trainer, tagger, tag_dictionary, corpus
    loaded_model: SequenceTagger = SequenceTagger.load(results_base_path / "final-model.pt")

    sentence = Sentence("I love Berlin")
    sentence_empty = Sentence("       ")

    loaded_model.predict(sentence)
    loaded_model.predict([sentence, sentence_empty])
    loaded_model.predict([sentence_empty])

    del loaded_model


@pytest.mark.integration
def test_train_load_use_tagger_disjunct_tags(results_base_path, tasks_base_path):
    corpus = flair.datasets.ColumnCorpus(
        data_folder=tasks_base_path / "fashion_disjunct",
        column_format={0: "text", 3: "ner"},
    )
    tag_dictionary = corpus.make_label_dictionary("ner")

    tagger: SequenceTagger = SequenceTagger(
        hidden_size=64,
        embeddings=turian_embeddings,
        tag_dictionary=tag_dictionary,
        tag_type="ner",
        use_crf=False,
        allow_unk_predictions=True,
    )

    # initialize trainer
    trainer: ModelTrainer = ModelTrainer(tagger, corpus)

    trainer.train(
        results_base_path,
        learning_rate=0.1,
        mini_batch_size=2,
        max_epochs=2,
        shuffle=False,
    )


@pytest.mark.integration
def test_train_load_use_tagger_large(results_base_path, tasks_base_path):
    corpus = flair.datasets.UD_ENGLISH().downsample(0.05)
    tag_dictionary = corpus.make_label_dictionary("pos")

    tagger: SequenceTagger = SequenceTagger(
        hidden_size=64,
        embeddings=turian_embeddings,
        tag_dictionary=tag_dictionary,
        tag_type="pos",
        use_crf=False,
    )

    # initialize trainer
    trainer: ModelTrainer = ModelTrainer(tagger, corpus)

    trainer.train(
        results_base_path,
        learning_rate=0.1,
        mini_batch_size=32,
        max_epochs=2,
        shuffle=False,
    )

    del trainer, tagger, tag_dictionary, corpus
    loaded_model: SequenceTagger = SequenceTagger.load(results_base_path / "final-model.pt")

    sentence = Sentence("I love Berlin")
    sentence_empty = Sentence("       ")

    loaded_model.predict(sentence)
    loaded_model.predict([sentence, sentence_empty])
    loaded_model.predict([sentence_empty])

    del loaded_model


@pytest.mark.integration
def test_train_load_use_tagger_flair_embeddings(results_base_path, tasks_base_path):
    corpus = flair.datasets.ColumnCorpus(data_folder=tasks_base_path / "fashion", column_format={0: "text", 3: "ner"})
    tag_dictionary = corpus.make_label_dictionary("ner")

    tagger: SequenceTagger = SequenceTagger(
        hidden_size=64,
        embeddings=flair_embeddings,
        tag_dictionary=tag_dictionary,
        tag_type="ner",
        use_crf=False,
    )

    # initialize trainer
    trainer: ModelTrainer = ModelTrainer(tagger, corpus)

    trainer.train(
        results_base_path,
        learning_rate=0.1,
        mini_batch_size=2,
        max_epochs=2,
        shuffle=False,
    )

    del trainer, tagger, tag_dictionary, corpus
    loaded_model: SequenceTagger = SequenceTagger.load(results_base_path / "final-model.pt")

    sentence = Sentence("I love Berlin")
    sentence_empty = Sentence("       ")

    loaded_model.predict(sentence)
    loaded_model.predict([sentence, sentence_empty])
    loaded_model.predict([sentence_empty])

    del loaded_model


@pytest.mark.integration
def test_train_load_use_tagger_adam(results_base_path, tasks_base_path):
    corpus = flair.datasets.ColumnCorpus(data_folder=tasks_base_path / "fashion", column_format={0: "text", 3: "ner"})
    tag_dictionary = corpus.make_label_dictionary("ner")

    tagger: SequenceTagger = SequenceTagger(
        hidden_size=64,
        embeddings=turian_embeddings,
        tag_dictionary=tag_dictionary,
        tag_type="ner",
        use_crf=False,
    )

    # initialize trainer
    trainer: ModelTrainer = ModelTrainer(tagger, corpus)

    trainer.train(
        results_base_path,
        learning_rate=0.1,
        mini_batch_size=2,
        max_epochs=2,
        shuffle=False,
        optimizer=Adam,
    )

    del trainer, tagger, tag_dictionary, corpus
    loaded_model: SequenceTagger = SequenceTagger.load(results_base_path / "final-model.pt")

    sentence = Sentence("I love Berlin")
    sentence_empty = Sentence("       ")

    loaded_model.predict(sentence)
    loaded_model.predict([sentence, sentence_empty])
    loaded_model.predict([sentence_empty])

    del loaded_model


@pytest.mark.integration
def test_train_load_use_tagger_multicorpus(results_base_path, tasks_base_path):
    corpus_1 = flair.datasets.ColumnCorpus(data_folder=tasks_base_path / "fashion", column_format={0: "text", 3: "ner"})
    corpus_2 = flair.datasets.NER_GERMAN_GERMEVAL(base_path=tasks_base_path).downsample(0.1)

    corpus = MultiCorpus([corpus_1, corpus_2])
    tag_dictionary = corpus.make_label_dictionary("ner")

    tagger: SequenceTagger = SequenceTagger(
        hidden_size=64,
        embeddings=turian_embeddings,
        tag_dictionary=tag_dictionary,
        tag_type="ner",
        use_crf=False,
        allow_unk_predictions=True,
    )

    # initialize trainer
    trainer: ModelTrainer = ModelTrainer(tagger, corpus)

    trainer.train(
        results_base_path,
        learning_rate=0.1,
        mini_batch_size=2,
        max_epochs=2,
        shuffle=False,
    )

    del trainer, tagger, corpus
    loaded_model: SequenceTagger = SequenceTagger.load(results_base_path / "final-model.pt")

    sentence = Sentence("I love Berlin")
    sentence_empty = Sentence("       ")

    loaded_model.predict(sentence)
    loaded_model.predict([sentence, sentence_empty])
    loaded_model.predict([sentence_empty])

    del loaded_model


@pytest.mark.integration
def test_train_resume_tagger(results_base_path, tasks_base_path):

    corpus_1 = flair.datasets.ColumnCorpus(data_folder=tasks_base_path / "fashion", column_format={0: "text", 3: "ner"})
    corpus_2 = flair.datasets.NER_GERMAN_GERMEVAL(base_path=tasks_base_path).downsample(0.1)

    corpus = MultiCorpus([corpus_1, corpus_2])
    tag_dictionary = corpus.make_label_dictionary("ner")

    model: SequenceTagger = SequenceTagger(
        hidden_size=64,
        embeddings=turian_embeddings,
        tag_dictionary=tag_dictionary,
        tag_type="ner",
        use_crf=False,
    )

    # train model for 2 epochs
    trainer = ModelTrainer(model, corpus)
    trainer.train(results_base_path, max_epochs=2, shuffle=False, checkpoint=True)

    del model

    # load the checkpoint model and train until epoch 4
    checkpoint_model = SequenceTagger.load(results_base_path / "checkpoint.pt")
    trainer.resume(model=checkpoint_model, max_epochs=4)

    # clean up results directory
    del trainer


@pytest.mark.integration
def test_find_learning_rate(results_base_path, tasks_base_path):
    corpus = flair.datasets.ColumnCorpus(data_folder=tasks_base_path / "fashion", column_format={0: "text", 3: "ner"})
    tag_dictionary = corpus.make_label_dictionary("ner")

    tagger: SequenceTagger = SequenceTagger(
        hidden_size=64,
        embeddings=turian_embeddings,
        tag_dictionary=tag_dictionary,
        tag_type="ner",
        use_crf=False,
    )

    # initialize trainer
    trainer: ModelTrainer = ModelTrainer(tagger, corpus)

    trainer.find_learning_rate(results_base_path, optimizer=SGD, iterations=5)

    del trainer, tagger, tag_dictionary, corpus



from typing import Tuple

import flair.datasets
from flair.data import Corpus
from flair.embeddings import DocumentRNNEmbeddings, WordEmbeddings
from flair.models.text_regression_model import TextRegressor

# from flair.trainers.trainer_regression import RegressorTrainer
from flair.trainers import ModelTrainer


def init(tasks_base_path) -> Tuple[Corpus, TextRegressor, ModelTrainer]:
    corpus = flair.datasets.ClassificationCorpus(tasks_base_path / "regression")

    glove_embedding: WordEmbeddings = WordEmbeddings("glove")
    document_embeddings: DocumentRNNEmbeddings = DocumentRNNEmbeddings(
        [glove_embedding], 128, 1, False, 64, False, False
    )

    model = TextRegressor(document_embeddings)

    trainer = ModelTrainer(model, corpus)

    return corpus, model, trainer


def test_labels_to_indices(tasks_base_path):
    corpus, model, trainer = init(tasks_base_path)

    result = model._labels_to_indices(corpus.train)

    for i in range(len(corpus.train)):
        expected = round(float(corpus.train[i].labels[0].value), 3)
        actual = round(float(result[i].item()), 3)

        assert expected == actual


# def test_trainer_evaluation(tasks_base_path):
#     corpus, model, trainer = init(tasks_base_path)
#
#     expected = model.evaluate(corpus.dev)
#
#     assert expected is not None


# def test_trainer_results(tasks_base_path):
#    corpus, model, trainer = init(tasks_base_path)

#    results = trainer.train("regression_train/", max_epochs=1)

#    assert results["test_score"] > 0
#    assert len(results["dev_loss_history"]) == 1
#    assert len(results["dev_score_history"]) == 1
#    assert len(results["train_loss_history"]) == 1



import logging
from math import floor
from pathlib import Path
from typing import List, Optional, Union

import numpy as np
import sklearn.metrics as metrics
import torch
import torch.nn as nn
from torch.nn.modules.loss import _Loss
from torch.utils.data.dataset import Dataset

import flair.embeddings
import flair.nn
from flair.data import Sentence, _iter_dataset
from flair.training_utils import MetricRegression, Result, store_embeddings

log = logging.getLogger("flair")


class WeightedMSELoss(_Loss):
    def __init__(
        self,
        regr_loss_step: float,
        size_average=None,
        reduce=None,
        reduction: str = "mean",
    ) -> None:
        super(WeightedMSELoss, self).__init__(size_average, reduce, reduction)
        self.regr_loss_step = regr_loss_step

    def forward(self, predictions: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        weight = 1 + self.regr_loss_step * target

        return (weight * ((predictions - target) ** 2)).mean()


class DistancePredictor(flair.nn.Model[Sentence]):
    """
    DistancePredictor
    Model to predict distance between two words given their embeddings, modeled either as a classification or a
    regression model. Takes (contextual) word embedding as input.
    The pair of word embeddings is passed through a linear layer that predicts their distance in a sentence.
    Note: When used for training the batch size must be set to 1!!!
    """

    def __init__(
        self,
        word_embeddings: flair.embeddings.TokenEmbeddings,
        max_distance: int = 20,
        beta: float = 1.0,
        loss_max_weight: float = 1,
        regression=False,
        regr_loss_step: float = 0,
    ):
        """
        Initializes a DistClassifier
        :param word_embeddings: embeddings used to embed each sentence
        .param max_distance: max dist between word pairs = number of predicted classes - 1
        :param beta: Parameter for F-beta score for evaluation and training annealing
        :param loss_max_weight: Only for classification: Since small distances between word pairs occur mor frequent it makes sense to give them less weight
        in the loss function. loss_max_weight will be used as the weight for the maximum distance and should be a number >=1
        The other weights decrease with equidistant steps from high to low distance.
        :param regression: if True the class does regression instead of classification
        :param regr_loss_step: if > 0, the MSE-Loss in regression will be weighted. Word pairs with
        distance 0 have weight 1. Then, as the distance increases, the weight in the loss function,
        increases step by step with size regr_loss_step
        """

        super(DistancePredictor, self).__init__()

        self.word_embeddings: flair.embeddings.TokenEmbeddings = word_embeddings

        self.beta = beta

        self.loss_max_weight = loss_max_weight

        self.regression = regression

        self.regr_loss_step = regr_loss_step

        if not regression:
            self.max_distance = max_distance

            # weights for loss function
            if self.loss_max_weight > 1:
                step = (self.loss_max_weight - 1) / self.max_distance

                weight_list = [1.0 + i * step for i in range(self.max_distance + 1)]

                self.loss_weights: Optional[torch.Tensor] = torch.FloatTensor(weight_list).to(flair.device)

            else:
                self.loss_weights = None

            # iput size is two times wordembedding size since we use pair of words as input
            # the output size is max_distance + 1, i.e. we allow 0,1,...,max_distance words between pairs
            self.decoder = nn.Linear(self.word_embeddings.embedding_length * 2, self.max_distance + 1)

            self.loss_function: _Loss = nn.CrossEntropyLoss(weight=self.loss_weights)

        # regression
        else:
            self.max_distance = 1000000

            # input size is two times word embedding size since we use pair of words as input
            # the output size is 1
            self.decoder = nn.Linear(self.word_embeddings.embedding_length * 2, 1)

            if regr_loss_step > 0:
                self.loss_function = WeightedMSELoss(regr_loss_step)
            else:
                self.loss_function = nn.MSELoss()

        nn.init.xavier_uniform_(self.decoder.weight)

        # auto-spawn on GPU if available
        self.to(flair.device)

    def label_type(self):
        return "distance"

    # forward allows only a single sentcence!!
    def forward(self, sentence: Sentence):

        # embed words of sentence
        self.word_embeddings.embed(sentence)

        # go through all pairs of words with a maximum number of max_distance in between
        numberOfWords = len(sentence)
        text_embedding_list = []
        # go through all pairs
        for i in range(numberOfWords):
            for j in range(i + 1, min(i + self.max_distance + 2, numberOfWords)):
                text_embedding_list.append(torch.cat((sentence[i].embedding, sentence[j].embedding)).unsqueeze(0))

        # 2-dim matrix whose rows are the embeddings of word pairs of the sentence
        text_embedding_tensor = torch.cat(text_embedding_list, 0).to(flair.device)

        label_scores = self.decoder(text_embedding_tensor)

        if self.regression:
            return label_scores.squeeze(1)

        return label_scores

    def _get_state_dict(self):
        model_state = {
            **super()._get_state_dict(),
            "word_embeddings": self.word_embeddings,
            "max_distance": self.max_distance,
            "beta": self.beta,
            "loss_max_weight": self.loss_max_weight,
            "regression": self.regression,
            "regr_loss_step": self.regr_loss_step,
        }
        return model_state

    @classmethod
    def _init_model_with_state_dict(cls, state, **kwargs):

        beta = 1.0 if "beta" not in state.keys() else state["beta"]
        weight = 1 if "loss_max_weight" not in state.keys() else state["loss_max_weight"]

        return super()._init_model_with_state_dict(
            state,
            word_embeddings=state["word_embeddings"],
            max_distance=state["max_distance"],
            beta=beta,
            loss_max_weight=weight,
            regression=state["regression"],
            regr_loss_step=state["regr_loss_step"],
            **kwargs,
        )

    # So far only one sentence allowed
    # If list of sentences is handed the function works with the first sentence of the list
    def forward_loss(self, data_points: Union[List[Sentence], Sentence]) -> torch.Tensor:

        if isinstance(data_points, list):  # first sentence
            data_points = data_points[0]

        if len(data_points) < 2:
            return torch.tensor([0.0], requires_grad=True)

        scores = self.forward(data_points)

        return self._calculate_loss(scores, data_points)

    # Assume data_points is a single sentence!!!
    # scores are the predictions for each word pair
    def _calculate_loss(self, scores, data_points):

        indices = []
        numberOfWords = len(data_points)

        # classification needs labels to be integers, regression needs labels to be float
        # this is due to the different loss functions
        if not self.regression:
            for i in range(numberOfWords):
                for j in range(i + 1, min(i + self.max_distance + 2, numberOfWords)):
                    indices.append(torch.LongTensor([j - i - 1]))  # distance between words
        else:
            for i in range(numberOfWords):
                for j in range(i + 1, min(i + self.max_distance + 2, numberOfWords)):
                    indices.append(torch.Tensor([j - i - 1]))  # distance between words

        labels = torch.cat(indices, 0).to(flair.device)

        return self.loss_function(scores, labels)

    # only single sentences as input
    def _forward_scores_and_loss(self, data_points: Union[List[Sentence], Sentence], return_loss=False):

        if isinstance(data_points, list):  # first sentence
            data_points = data_points[0]

        scores = self.forward(data_points)

        loss = None
        if return_loss:
            loss = self._calculate_loss(scores, data_points)

        return scores, loss

    def evaluate(
        self,
        data_points: Union[List[Sentence], Dataset],
        gold_label_type: str,
        out_path: Union[str, Path] = None,
        embedding_storage_mode: str = "none",
        mini_batch_size: int = 1,  # unnecessary, but trainer.train calls evaluate with this parameter
        num_workers: Optional[int] = 8,
        **kwargs,
    ) -> Result:

        if isinstance(data_points, Dataset):
            data_points = list(_iter_dataset(data_points))

        if self.regression:
            return self.evaluate_regression(
                sentences=data_points,
                out_path=out_path,
                embedding_storage_mode=embedding_storage_mode,
            )

        return self.evaluate_classification(
            sentences=data_points,
            out_path=out_path,
            embedding_storage_mode=embedding_storage_mode,
        )

    def evaluate_regression(
        self,
        sentences: List[Sentence],
        out_path: Union[str, Path] = None,
        embedding_storage_mode: str = "none",
    ) -> Result:

        with torch.no_grad():

            buckets = [0 for _ in range(11)]

            eval_loss = torch.zeros(1, device=flair.device)

            metric = MetricRegression("Evaluation")

            lines: List[str] = []

            max_dist_plus_one = max([len(sent) for sent in sentences]) - 1

            num_occurences = [0 for _ in range(max_dist_plus_one)]

            cumulated_values = [0 for _ in range(max_dist_plus_one)]

            for sentence in sentences:

                if len(sentence) < 2:  # we need at least 2 words per sentence
                    continue

                scores, loss = self._forward_scores_and_loss(sentence, return_loss=True)

                predictions = scores.tolist()

                # gold labels
                true_values_for_sentence = []
                numberOfPairs = 0
                numberOfWords = len(sentence)
                lines.append(sentence.to_tokenized_string() + "\n")
                for i in range(numberOfWords):
                    for j in range(i + 1, min(i + self.max_distance + 2, numberOfWords)):
                        true_dist = j - i - 1
                        pred = predictions[numberOfPairs]

                        true_values_for_sentence.append(true_dist)

                        # for output text file
                        eval_line = f"({i},{j})\t{true_dist}\t{pred:.2f}\n"
                        lines.append(eval_line)

                        # for buckets
                        error = abs(true_dist - pred)
                        if error >= 10:
                            buckets[10] += 1
                        else:
                            buckets[floor(error)] += 1

                        # for average prediction
                        num_occurences[true_dist] += 1
                        cumulated_values[true_dist] += pred

                        numberOfPairs += 1

                eval_loss += loss / numberOfPairs

                metric.true.extend(true_values_for_sentence)
                metric.pred.extend(predictions)

                store_embeddings([sentence], embedding_storage_mode)

            eval_loss /= len(sentences)  # w.r.t self.loss

            # add some statistics to the output
            eval_line = f"Number of Sentences: {len(sentences)}\nBuckets:\n | 0-1 | 1-2 | 2-3 | 3-4 | 4-5 | 5-6 | 6-7 | 7-8 | 8-9 | 9-10 | >10 |\n"
            lines.append(eval_line)
            eval_line = "| {} | {} | {} | {} | {} | {} | {} | {} | {} | {} | {} |".format(
                buckets[0],
                buckets[1],
                buckets[2],
                buckets[3],
                buckets[4],
                buckets[5],
                buckets[6],
                buckets[7],
                buckets[8],
                buckets[9],
                buckets[10],
            )
            lines.append(eval_line)
            lines.append("\nAverage predicted values per distance:\n")
            eval_line = ""
            for i in range(max_dist_plus_one):
                eval_line += str(i) + ": " + f"{cumulated_values[i] / num_occurences[i]:.2f}" + " "
                if i != 0 and i % 15 == 0:
                    eval_line += "\n"

            lines.append(eval_line)

            if out_path is not None:
                with open(out_path, "w", encoding="utf-8") as outfile:
                    outfile.write("".join(lines))

            log_line = f"{metric.mean_squared_error()}\t{metric.spearmanr()}\t{metric.pearsonr()}"
            log_header = "MSE\tSPEARMAN\tPEARSON"

            detailed_result = (
                f"AVG: mse: {metric.mean_squared_error():.4f} - "
                f"mae: {metric.mean_absolute_error():.4f} - "
                f"pearson: {metric.pearsonr():.4f} - "
                f"spearman: {metric.spearmanr():.4f}"
            )

            result: Result = Result(metric.pearsonr(), log_header, log_line, detailed_result, loss=eval_loss.item())

            return result

    def evaluate_classification(
        self,
        sentences: List[Sentence],
        out_path: Union[str, Path] = None,
        embedding_storage_mode: str = "none",
    ) -> Result:

        # use scikit-learn to evaluate
        y_true = []
        y_pred = []

        with torch.no_grad():
            eval_loss = torch.zeros(1, device=flair.device)

            lines: List[str] = []
            # we iterate over each sentence, instead of batches
            for sentence in sentences:

                if len(sentence) < 2:  # we need at least 2 words per sentence
                    continue

                scores, loss = self._forward_scores_and_loss(sentence, return_loss=True)

                # get single labels from scores
                predictions = [self._get_single_label(s) for s in scores]

                # gold labels
                true_values_for_sentence = []
                numberOfPairs = 0
                numberOfWords = len(sentence)
                lines.append(sentence.to_tokenized_string() + "\n")
                for i in range(numberOfWords):
                    for j in range(i + 1, min(i + self.max_distance + 2, numberOfWords)):
                        true_values_for_sentence.append(j - i - 1)

                        # for output text file
                        eval_line = "({},{})\t{}\t{}\n".format(i, j, j - i - 1, predictions[numberOfPairs])
                        lines.append(eval_line)

                        numberOfPairs += 1

                eval_loss += loss / numberOfPairs  # add average loss of word pairs

                for prediction_for_sentence, true_value_for_sentence in zip(predictions, true_values_for_sentence):
                    # hot one vector of true value
                    y_true_instance = np.zeros(self.max_distance + 1, dtype=int)
                    y_true_instance[true_value_for_sentence] = 1
                    y_true.append(y_true_instance.tolist())

                    # hot one vector of predicted value
                    y_pred_instance = np.zeros(self.max_distance + 1, dtype=int)
                    y_pred_instance[prediction_for_sentence] = 1
                    y_pred.append(y_pred_instance.tolist())

                # speichert embeddings, falls embedding_storage!= 'None'
                store_embeddings([sentence], embedding_storage_mode)

            if out_path is not None:
                with open(out_path, "w", encoding="utf-8") as outfile:
                    outfile.write("".join(lines))

            # make "classification report"
            target_names = []  # liste aller labels, ins unserem Fall
            for i in range(self.max_distance + 1):
                target_names.append(str(i))
            classification_report = metrics.classification_report(
                y_true, y_pred, digits=4, target_names=target_names, zero_division=0
            )

            # get scores
            micro_f_score = round(
                metrics.fbeta_score(y_true, y_pred, beta=self.beta, average="micro", zero_division=0),
                4,
            )
            accuracy_score = round(metrics.accuracy_score(y_true, y_pred), 4)
            macro_f_score = round(
                metrics.fbeta_score(y_true, y_pred, beta=self.beta, average="macro", zero_division=0),
                4,
            )
            # precision_score = round(metrics.precision_score(y_true, y_pred, average='macro', zero_division=0), 4)
            # recall_score = round(metrics.recall_score(y_true, y_pred, average='macro', zero_division=0), 4)

            detailed_result = (
                "\nResults:"
                f"\n- F-score (micro) {micro_f_score}"
                f"\n- F-score (macro) {macro_f_score}"
                f"\n- Accuracy {accuracy_score}"
                "\n\nBy class:\n" + classification_report
            )

            # line for log file
            log_header = "ACCURACY"
            log_line = f"\t{accuracy_score}"
            eval_loss /= len(sentences)

            result = Result(
                main_score=micro_f_score,
                log_line=log_line,
                log_header=log_header,
                detailed_results=detailed_result,
                loss=eval_loss.item(),
            )

            return result

    @staticmethod
    def _filter_empty_sentences(sentences: List[Sentence]) -> List[Sentence]:
        filtered_sentences = [sentence for sentence in sentences if sentence.tokens]
        if len(sentences) != len(filtered_sentences):
            log.warning("Ignore {} sentence(s) with no tokens.".format(len(sentences) - len(filtered_sentences)))
        return filtered_sentences

    def _get_single_label(self, label_scores):  # -> List[Label]:
        softmax = torch.nn.functional.softmax(label_scores, dim=0)
        conf, idx = torch.max(softmax, 0)

        return idx.item()

    def __str__(self):
        return (
            super(flair.nn.Model, self).__str__().rstrip(")")
            + f"  (beta): {self.beta}\n"
            + f"  (loss_max_weight): {self.loss_max_weight}\n"
            + f"  (max_distance) {self.max_distance}\n)"
        )



import torch


class LockedDropout(torch.nn.Module):
    """
    Implementation of locked (or variational) dropout. Randomly drops out entire parameters in embedding space.
    """

    def __init__(self, dropout_rate=0.5, batch_first=True, inplace=False):
        super(LockedDropout, self).__init__()
        self.dropout_rate = dropout_rate
        self.batch_first = batch_first
        self.inplace = inplace

    def forward(self, x):
        if not self.training or not self.dropout_rate:
            return x

        if not self.batch_first:
            m = x.data.new(1, x.size(1), x.size(2)).bernoulli_(1 - self.dropout_rate)
        else:
            m = x.data.new(x.size(0), 1, x.size(2)).bernoulli_(1 - self.dropout_rate)

        mask = torch.autograd.Variable(m, requires_grad=False) / (1 - self.dropout_rate)
        mask = mask.expand_as(x)
        return mask * x

    def extra_repr(self):
        inplace_str = ", inplace" if self.inplace else ""
        return "p={}{}".format(self.dropout_rate, inplace_str)


class WordDropout(torch.nn.Module):
    """
    Implementation of word dropout. Randomly drops out entire words (or characters) in embedding space.
    """

    def __init__(self, dropout_rate=0.05, inplace=False):
        super(WordDropout, self).__init__()
        self.dropout_rate = dropout_rate
        self.inplace = inplace

    def forward(self, x):
        if not self.training or not self.dropout_rate:
            return x

        m = x.data.new(x.size(0), x.size(1), 1).bernoulli_(1 - self.dropout_rate)

        mask = torch.autograd.Variable(m, requires_grad=False)
        return mask * x

    def extra_repr(self):
        inplace_str = ", inplace" if self.inplace else ""
        return "p={}{}".format(self.dropout_rate, inplace_str)



import pytest

import flair.datasets
from flair.data import Sentence
from flair.embeddings import DocumentRNNEmbeddings, FlairEmbeddings, WordEmbeddings
from flair.models import TextClassifier
from flair.samplers import ImbalancedClassificationDatasetSampler
from flair.trainers import ModelTrainer

turian_embeddings = WordEmbeddings("turian")
flair_embeddings = FlairEmbeddings("news-forward-fast")
document_embeddings: DocumentRNNEmbeddings = DocumentRNNEmbeddings([turian_embeddings], 128, 1, False, 64, False, False)


@pytest.mark.integration
def test_load_use_classifier():
    loaded_model: TextClassifier = TextClassifier.load("sentiment")

    sentence = Sentence("I love Berlin")
    sentence_empty = Sentence("       ")

    loaded_model.predict(sentence)
    loaded_model.predict([sentence, sentence_empty])
    loaded_model.predict([sentence_empty])
    del loaded_model

    sentence.clear_embeddings()
    sentence_empty.clear_embeddings()


@pytest.mark.integration
def test_train_load_use_classifier(results_base_path, tasks_base_path):
    corpus = flair.datasets.ClassificationCorpus(tasks_base_path / "imdb", label_type="topic")
    label_dict = corpus.make_label_dictionary(label_type="topic")

    model: TextClassifier = TextClassifier(
        document_embeddings=document_embeddings,
        label_dictionary=label_dict,
        label_type="topic",
        multi_label=False,
    )

    trainer = ModelTrainer(model, corpus)
    trainer.train(results_base_path, max_epochs=2, shuffle=False)

    sentence = Sentence("Berlin is a really nice city.")

    model.predict(sentence)

    for label in sentence.labels:
        assert label.value is not None
        assert 0.0 <= label.score <= 1.0
        assert type(label.score) is float

    del trainer, model, corpus
    loaded_model = TextClassifier.load(results_base_path / "final-model.pt")

    sentence = Sentence("I love Berlin")
    sentence_empty = Sentence("       ")

    loaded_model.predict(sentence)
    loaded_model.predict([sentence, sentence_empty])
    loaded_model.predict([sentence_empty])

    del loaded_model


@pytest.mark.integration
def test_train_load_use_classifier_with_sampler(results_base_path, tasks_base_path):
    corpus = flair.datasets.ClassificationCorpus(tasks_base_path / "imdb", label_type="topic")
    label_dict = corpus.make_label_dictionary(label_type="topic")

    model: TextClassifier = TextClassifier(
        document_embeddings=document_embeddings,
        label_dictionary=label_dict,
        label_type="topic",
        multi_label=False,
    )

    trainer = ModelTrainer(model, corpus)
    trainer.train(
        results_base_path,
        max_epochs=2,
        shuffle=False,
        sampler=ImbalancedClassificationDatasetSampler,
    )

    sentence = Sentence("Berlin is a really nice city.")
    model.predict(sentence)

    for label in sentence.labels:
        assert label.value is not None
        assert 0.0 <= label.score <= 1.0
        assert type(label.score) is float

    del trainer, model, corpus
    loaded_model = TextClassifier.load(results_base_path / "final-model.pt")

    sentence = Sentence("I love Berlin")
    sentence_empty = Sentence("       ")

    loaded_model.predict(sentence)
    loaded_model.predict([sentence, sentence_empty])
    loaded_model.predict([sentence_empty])

    del loaded_model


@pytest.mark.integration
def test_train_load_use_classifier_with_prob(results_base_path, tasks_base_path):
    corpus = flair.datasets.ClassificationCorpus(tasks_base_path / "imdb", label_type="topic")
    label_dict = corpus.make_label_dictionary(label_type="topic")

    model: TextClassifier = TextClassifier(
        document_embeddings=document_embeddings,
        label_dictionary=label_dict,
        label_type="topic",
        multi_label=False,
    )

    trainer = ModelTrainer(model, corpus)
    trainer.train(results_base_path, max_epochs=2, shuffle=False)

    sentence = Sentence("Berlin is a really nice city.")

    model.predict(sentence, return_probabilities_for_all_classes=True)

    assert len(sentence.labels) > 1

    for label in sentence.labels:
        assert label.value is not None
        assert 0.0 <= label.score <= 1.0
        assert type(label.score) is float

    del trainer, model, corpus
    loaded_model = TextClassifier.load(results_base_path / "final-model.pt")

    sentence = Sentence("I love Berlin")
    sentence_empty = Sentence("       ")

    loaded_model.predict(sentence, return_probabilities_for_all_classes=True)
    loaded_model.predict([sentence, sentence_empty], return_probabilities_for_all_classes=True)
    loaded_model.predict([sentence_empty], return_probabilities_for_all_classes=True)

    del loaded_model


@pytest.mark.integration
def test_train_load_use_classifier_multi_label(results_base_path, tasks_base_path):
    corpus = flair.datasets.ClassificationCorpus(tasks_base_path / "multi_class", label_type="topic")
    label_dict = corpus.make_label_dictionary(label_type="topic")

    model: TextClassifier = TextClassifier(
        document_embeddings=document_embeddings,
        label_dictionary=label_dict,
        label_type="topic",
        multi_label=True,
    )

    trainer = ModelTrainer(model, corpus)
    trainer.train(
        results_base_path,
        mini_batch_size=1,
        max_epochs=20,
        shuffle=False,
        checkpoint=False,
        train_with_test=True,
        train_with_dev=True,
    )

    sentence = Sentence("apple tv")

    model.predict(sentence)

    assert "apple" in [label.value for label in sentence.labels]
    assert "tv" in [label.value for label in sentence.labels]

    for label in sentence.labels:
        print(label)
        assert label.value is not None
        assert 0.0 <= label.score <= 1.0
        assert type(label.score) is float

    del trainer, model, corpus
    loaded_model = TextClassifier.load(results_base_path / "final-model.pt")

    sentence = Sentence("apple tv")

    loaded_model.predict(sentence)

    assert "apple" in [label.value for label in sentence.labels]
    assert "tv" in [label.value for label in sentence.labels]

    for label in sentence.labels:
        assert label.value is not None
        assert 0.0 <= label.score <= 1.0
        assert type(label.score) is float

    sentence = Sentence("I love Berlin")
    sentence_empty = Sentence("       ")

    loaded_model.predict(sentence)
    loaded_model.predict([sentence, sentence_empty])
    loaded_model.predict([sentence_empty])

    del loaded_model


@pytest.mark.integration
def test_train_load_use_classifier_flair(results_base_path, tasks_base_path):
    corpus = flair.datasets.ClassificationCorpus(tasks_base_path / "imdb", label_type="topic")
    label_dict = corpus.make_label_dictionary(label_type="topic")

    flair_document_embeddings: DocumentRNNEmbeddings = DocumentRNNEmbeddings(
        [flair_embeddings], 128, 1, False, 64, False, False
    )

    model: TextClassifier = TextClassifier(
        document_embeddings=flair_document_embeddings,
        label_dictionary=label_dict,
        label_type="topic",
        multi_label=False,
    )

    trainer = ModelTrainer(model, corpus)
    trainer.train(results_base_path, max_epochs=2, shuffle=False)

    sentence = Sentence("Berlin is a really nice city.")

    model.predict(sentence)

    for label in sentence.labels:
        assert label.value is not None
        assert 0.0 <= label.score <= 1.0
        assert type(label.score) is float

    del trainer, model, corpus, flair_document_embeddings
    loaded_model = TextClassifier.load(results_base_path / "final-model.pt")

    sentence = Sentence("I love Berlin")
    sentence_empty = Sentence("       ")

    loaded_model.predict(sentence)
    loaded_model.predict([sentence, sentence_empty])
    loaded_model.predict([sentence_empty])

    del loaded_model


@pytest.mark.integration
def test_train_resume_classifier(results_base_path, tasks_base_path):
    corpus = flair.datasets.ClassificationCorpus(tasks_base_path / "imdb", label_type="topic")
    label_dict = corpus.make_label_dictionary(label_type="topic")

    model = TextClassifier(
        document_embeddings=document_embeddings,
        label_dictionary=label_dict,
        multi_label=False,
        label_type="topic",
    )

    # train model for 2 epochs
    trainer = ModelTrainer(model, corpus)
    trainer.train(results_base_path, max_epochs=2, shuffle=False, checkpoint=True)

    del model

    # load the checkpoint model and train until epoch 4
    checkpoint_model = TextClassifier.load(results_base_path / "checkpoint.pt")
    with pytest.warns(UserWarning):
        trainer.resume(model=checkpoint_model, max_epochs=4)

    del trainer



# Expose base classses
from .base import DataLoader, FlairDatapointDataset, MongoDataset, StringDataset

# Expose all biomedical data sets used for the evaluation of BioBERT
# -
# -
# -
# -
# Expose all biomedical data sets using the HUNER splits
# Expose all biomedical data sets
from .biomedical import (
    ANAT_EM,
    AZDZ,
    BC2GM,
    BIO_INFER,
    BIOBERT_CHEMICAL_BC4CHEMD,
    BIOBERT_CHEMICAL_BC5CDR,
    BIOBERT_DISEASE_BC5CDR,
    BIOBERT_DISEASE_NCBI,
    BIOBERT_GENE_BC2GM,
    BIOBERT_GENE_JNLPBA,
    BIOBERT_SPECIES_LINNAEUS,
    BIOBERT_SPECIES_S800,
    BIONLP2013_CG,
    BIONLP2013_PC,
    BIOSEMANTICS,
    CDR,
    CELL_FINDER,
    CEMP,
    CHEMDNER,
    CLL,
    CRAFT,
    CRAFT_V4,
    DECA,
    FSU,
    GELLUS,
    GPRO,
    HUNER_CELL_LINE,
    HUNER_CELL_LINE_CELL_FINDER,
    HUNER_CELL_LINE_CLL,
    HUNER_CELL_LINE_GELLUS,
    HUNER_CELL_LINE_JNLPBA,
    HUNER_CHEMICAL,
    HUNER_CHEMICAL_CDR,
    HUNER_CHEMICAL_CEMP,
    HUNER_CHEMICAL_CHEBI,
    HUNER_CHEMICAL_CHEMDNER,
    HUNER_CHEMICAL_CRAFT_V4,
    HUNER_CHEMICAL_SCAI,
    HUNER_DISEASE,
    HUNER_DISEASE_CDR,
    HUNER_DISEASE_MIRNA,
    HUNER_DISEASE_NCBI,
    HUNER_DISEASE_PDR,
    HUNER_DISEASE_SCAI,
    HUNER_DISEASE_VARIOME,
    HUNER_GENE,
    HUNER_GENE_BC2GM,
    HUNER_GENE_BIO_INFER,
    HUNER_GENE_CELL_FINDER,
    HUNER_GENE_CHEBI,
    HUNER_GENE_CRAFT_V4,
    HUNER_GENE_DECA,
    HUNER_GENE_FSU,
    HUNER_GENE_GPRO,
    HUNER_GENE_IEPA,
    HUNER_GENE_JNLPBA,
    HUNER_GENE_LOCTEXT,
    HUNER_GENE_MIRNA,
    HUNER_GENE_OSIRIS,
    HUNER_GENE_VARIOME,
    HUNER_SPECIES,
    HUNER_SPECIES_CELL_FINDER,
    HUNER_SPECIES_CHEBI,
    HUNER_SPECIES_CRAFT_V4,
    HUNER_SPECIES_LINNEAUS,
    HUNER_SPECIES_LOCTEXT,
    HUNER_SPECIES_MIRNA,
    HUNER_SPECIES_S800,
    HUNER_SPECIES_VARIOME,
    IEPA,
    JNLPBA,
    LINNEAUS,
    LOCTEXT,
    MIRNA,
    NCBI_DISEASE,
    OSIRIS,
    PDR,
    S800,
    SCAI_CHEMICALS,
    SCAI_DISEASE,
    VARIOME,
)

# Expose all document classification datasets
from .document_classification import (
    AMAZON_REVIEWS,
    COMMUNICATIVE_FUNCTIONS,
    GERMEVAL_2018_OFFENSIVE_LANGUAGE,
    GLUE_COLA,
    GO_EMOTIONS,
    IMDB,
    NEWSGROUPS,
    SENTEVAL_CR,
    SENTEVAL_MPQA,
    SENTEVAL_MR,
    SENTEVAL_SST_BINARY,
    SENTEVAL_SST_GRANULAR,
    SENTEVAL_SUBJ,
    SENTIMENT_140,
    STACKOVERFLOW,
    TREC_6,
    TREC_50,
    WASSA_ANGER,
    WASSA_FEAR,
    WASSA_JOY,
    WASSA_SADNESS,
    YAHOO_ANSWERS,
    ClassificationCorpus,
    ClassificationDataset,
    CSVClassificationCorpus,
    CSVClassificationDataset,
)

# word sense disambiguation
# Expose all entity linking datasets
from .entity_linking import (
    NEL_ENGLISH_AIDA,
    NEL_ENGLISH_AQUAINT,
    NEL_ENGLISH_IITB,
    NEL_ENGLISH_REDDIT,
    NEL_ENGLISH_TWEEKI,
    NEL_GERMAN_HIPE,
    WSD_MASC,
    WSD_OMSTI,
    WSD_RAGANATO_ALL,
    WSD_SEMCOR,
    WSD_TRAINOMATIC,
    WSD_UFSAC,
    WSD_WORDNET_GLOSS_TAGGED,
)

# Expose all relation extraction datasets
from .relation_extraction import (
    RE_ENGLISH_CONLL04,
    RE_ENGLISH_DRUGPROT,
    RE_ENGLISH_SEMEVAL2010,
    RE_ENGLISH_TACRED,
)

# universal proposition banks
# keyphrase detection datasets
# other NER datasets
# standard NER datasets
# Expose all sequence labeling datasets
from .sequence_labeling import (
    BIOSCOPE,
    CONLL_03,
    CONLL_03_DUTCH,
    CONLL_03_GERMAN,
    CONLL_03_SPANISH,
    CONLL_2000,
    KEYPHRASE_INSPEC,
    KEYPHRASE_SEMEVAL2010,
    KEYPHRASE_SEMEVAL2017,
    NER_ARABIC_ANER,
    NER_ARABIC_AQMAR,
    NER_BASQUE,
    NER_CHINESE_WEIBO,
    NER_DANISH_DANE,
    NER_ENGLISH_MOVIE_COMPLEX,
    NER_ENGLISH_MOVIE_SIMPLE,
    NER_ENGLISH_PERSON,
    NER_ENGLISH_RESTAURANT,
    NER_ENGLISH_SEC_FILLINGS,
    NER_ENGLISH_STACKOVERFLOW,
    NER_ENGLISH_TWITTER,
    NER_ENGLISH_WEBPAGES,
    NER_ENGLISH_WIKIGOLD,
    NER_ENGLISH_WNUT_2020,
    NER_FINNISH,
    NER_GERMAN_BIOFID,
    NER_GERMAN_EUROPARL,
    NER_GERMAN_GERMEVAL,
    NER_GERMAN_LEGAL,
    NER_GERMAN_POLITICS,
    NER_HIPE_2022,
    NER_HUNGARIAN,
    NER_ICELANDIC,
    NER_JAPANESE,
    NER_MASAKHANE,
    NER_MULTI_CONER,
    NER_MULTI_WIKIANN,
    NER_MULTI_WIKINER,
    NER_MULTI_XTREME,
    NER_SWEDISH,
    NER_TURKU,
    UP_CHINESE,
    UP_ENGLISH,
    UP_FINNISH,
    UP_FRENCH,
    UP_GERMAN,
    UP_ITALIAN,
    UP_SPANISH,
    UP_SPANISH_ANCORA,
    WNUT_17,
    ColumnCorpus,
    ColumnDataset,
)

# Expose all text-image datasets
from .text_image import FeideggerCorpus, FeideggerDataset

# Expose all text-text datasets
from .text_text import (
    GLUE_MNLI,
    GLUE_MRPC,
    GLUE_QNLI,
    GLUE_QQP,
    GLUE_RTE,
    GLUE_WNLI,
    SUPERGLUE_RTE,
    DataPairCorpus,
    DataPairDataset,
    OpusParallelCorpus,
    ParallelTextCorpus,
    ParallelTextDataset,
)

# Expose all treebanks
from .treebanks import (
    UD_AFRIKAANS,
    UD_ANCIENT_GREEK,
    UD_ARABIC,
    UD_ARMENIAN,
    UD_BASQUE,
    UD_BELARUSIAN,
    UD_BULGARIAN,
    UD_CATALAN,
    UD_CHINESE,
    UD_COPTIC,
    UD_CROATIAN,
    UD_CZECH,
    UD_DANISH,
    UD_DUTCH,
    UD_ENGLISH,
    UD_ESTONIAN,
    UD_FAROESE,
    UD_FINNISH,
    UD_FRENCH,
    UD_GALICIAN,
    UD_GERMAN,
    UD_GERMAN_HDT,
    UD_GOTHIC,
    UD_GREEK,
    UD_HEBREW,
    UD_HINDI,
    UD_INDONESIAN,
    UD_IRISH,
    UD_ITALIAN,
    UD_JAPANESE,
    UD_KAZAKH,
    UD_KOREAN,
    UD_LATIN,
    UD_LATVIAN,
    UD_LITHUANIAN,
    UD_LIVVI,
    UD_MALTESE,
    UD_MARATHI,
    UD_NORTH_SAMI,
    UD_NORWEGIAN,
    UD_OLD_CHURCH_SLAVONIC,
    UD_OLD_FRENCH,
    UD_PERSIAN,
    UD_POLISH,
    UD_PORTUGUESE,
    UD_ROMANIAN,
    UD_RUSSIAN,
    UD_SERBIAN,
    UD_SLOVAK,
    UD_SLOVENIAN,
    UD_SPANISH,
    UD_SWEDISH,
    UD_TURKISH,
    UD_WOLOF,
    UniversalDependenciesCorpus,
    UniversalDependenciesDataset,
)

__all__ = [
    "DataLoader",
    "FlairDatapointDataset",
    "MongoDataset",
    "StringDataset",
    "ANAT_EM",
    "AZDZ",
    "BC2GM",
    "BIO_INFER",
    "BIOBERT_CHEMICAL_BC4CHEMD",
    "BIOBERT_CHEMICAL_BC5CDR",
    "BIOBERT_DISEASE_BC5CDR",
    "BIOBERT_DISEASE_NCBI",
    "BIOBERT_GENE_BC2GM",
    "BIOBERT_GENE_JNLPBA",
    "BIOBERT_SPECIES_LINNAEUS",
    "BIOBERT_SPECIES_S800",
    "BIONLP2013_CG",
    "BIONLP2013_PC",
    "BIOSEMANTICS",
    "CDR",
    "CELL_FINDER",
    "CEMP",
    "CHEMDNER",
    "CLL",
    "CRAFT",
    "CRAFT_V4",
    "DECA",
    "FSU",
    "GELLUS",
    "GPRO",
    "HUNER_CELL_LINE",
    "HUNER_CELL_LINE_CELL_FINDER",
    "HUNER_CELL_LINE_CLL",
    "HUNER_CELL_LINE_GELLUS",
    "HUNER_CELL_LINE_JNLPBA",
    "HUNER_CHEMICAL",
    "HUNER_CHEMICAL_CDR",
    "HUNER_CHEMICAL_CEMP",
    "HUNER_CHEMICAL_CHEBI",
    "HUNER_CHEMICAL_CHEMDNER",
    "HUNER_CHEMICAL_CRAFT_V4",
    "HUNER_CHEMICAL_SCAI",
    "HUNER_DISEASE",
    "HUNER_DISEASE_CDR",
    "HUNER_DISEASE_MIRNA",
    "HUNER_DISEASE_NCBI",
    "HUNER_DISEASE_PDR",
    "HUNER_DISEASE_SCAI",
    "HUNER_DISEASE_VARIOME",
    "HUNER_GENE",
    "HUNER_GENE_BC2GM",
    "HUNER_GENE_BIO_INFER",
    "HUNER_GENE_CELL_FINDER",
    "HUNER_GENE_CHEBI",
    "HUNER_GENE_CRAFT_V4",
    "HUNER_GENE_DECA",
    "HUNER_GENE_FSU",
    "HUNER_GENE_GPRO",
    "HUNER_GENE_IEPA",
    "HUNER_GENE_JNLPBA",
    "HUNER_GENE_LOCTEXT",
    "HUNER_GENE_MIRNA",
    "HUNER_GENE_OSIRIS",
    "HUNER_GENE_VARIOME",
    "HUNER_SPECIES",
    "HUNER_SPECIES_CELL_FINDER",
    "HUNER_SPECIES_CHEBI",
    "HUNER_SPECIES_CRAFT_V4",
    "HUNER_SPECIES_LINNEAUS",
    "HUNER_SPECIES_LOCTEXT",
    "HUNER_SPECIES_MIRNA",
    "HUNER_SPECIES_S800",
    "HUNER_SPECIES_VARIOME",
    "IEPA",
    "JNLPBA",
    "LINNEAUS",
    "LOCTEXT",
    "MIRNA",
    "NCBI_DISEASE",
    "OSIRIS",
    "PDR",
    "S800",
    "SCAI_CHEMICALS",
    "SCAI_DISEASE",
    "VARIOME",
    "AMAZON_REVIEWS",
    "COMMUNICATIVE_FUNCTIONS",
    "GERMEVAL_2018_OFFENSIVE_LANGUAGE",
    "GLUE_COLA",
    "GO_EMOTIONS",
    "IMDB",
    "NEWSGROUPS",
    "STACKOVERFLOW",
    "SENTEVAL_CR",
    "SENTEVAL_MPQA",
    "SENTEVAL_MR",
    "SENTEVAL_SST_BINARY",
    "SENTEVAL_SST_GRANULAR",
    "SENTEVAL_SUBJ",
    "SENTIMENT_140",
    "TREC_6",
    "TREC_50",
    "WASSA_ANGER",
    "WASSA_FEAR",
    "WASSA_JOY",
    "WASSA_SADNESS",
    "YAHOO_ANSWERS",
    "ClassificationCorpus",
    "ClassificationDataset",
    "CSVClassificationCorpus",
    "CSVClassificationDataset",
    "NEL_ENGLISH_AIDA",
    "NEL_ENGLISH_AQUAINT",
    "NEL_ENGLISH_IITB",
    "NEL_ENGLISH_REDDIT",
    "NEL_ENGLISH_TWEEKI",
    "NEL_GERMAN_HIPE",
    "WSD_MASC",
    "WSD_OMSTI",
    "WSD_RAGANATO_ALL",
    "WSD_SEMCOR",
    "WSD_TRAINOMATIC",
    "WSD_UFSAC",
    "WSD_WORDNET_GLOSS_TAGGED",
    "EntityLinkingCorpus",
    "RE_ENGLISH_CONLL04",
    "RE_ENGLISH_DRUGPROT",
    "RE_ENGLISH_SEMEVAL2010",
    "RE_ENGLISH_TACRED",
    "BIOSCOPE",
    "CONLL_03",
    "CONLL_03_DUTCH",
    "CONLL_03_GERMAN",
    "CONLL_03_SPANISH",
    "CONLL_2000",
    "KEYPHRASE_INSPEC",
    "KEYPHRASE_SEMEVAL2010",
    "KEYPHRASE_SEMEVAL2017",
    "NER_ARABIC_ANER",
    "NER_ARABIC_AQMAR",
    "NER_BASQUE",
    "NER_CHINESE_WEIBO",
    "NER_DANISH_DANE",
    "NER_ENGLISH_MOVIE_COMPLEX",
    "NER_ENGLISH_MOVIE_SIMPLE",
    "NER_ENGLISH_PERSON",
    "NER_ENGLISH_RESTAURANT",
    "NER_ENGLISH_SEC_FILLINGS",
    "NER_ENGLISH_STACKOVERFLOW",
    "NER_ENGLISH_TWITTER",
    "NER_ENGLISH_WEBPAGES",
    "NER_ENGLISH_WIKIGOLD",
    "NER_ENGLISH_WNUT_2020",
    "NER_FINNISH",
    "NER_GERMAN_BIOFID",
    "NER_GERMAN_EUROPARL",
    "NER_GERMAN_GERMEVAL",
    "NER_GERMAN_LEGAL",
    "NER_GERMAN_POLITICS",
    "NER_HIPE_2022",
    "NER_HUNGARIAN",
    "NER_ICELANDIC",
    "NER_JAPANESE",
    "NER_MASAKHANE",
    "NER_MULTI_WIKIANN",
    "NER_MULTI_WIKINER",
    "NER_MULTI_XTREME",
    "NER_SWEDISH",
    "NER_TURKU",
    "UP_CHINESE",
    "UP_ENGLISH",
    "UP_FINNISH",
    "UP_FRENCH",
    "UP_GERMAN",
    "UP_ITALIAN",
    "UP_SPANISH",
    "UP_SPANISH_ANCORA",
    "WNUT_17",
    "ColumnCorpus",
    "ColumnDataset",
    "NER_MULTI_CONER",
    "FeideggerCorpus",
    "FeideggerDataset",
    "GLUE_MNLI",
    "GLUE_MRPC",
    "GLUE_QNLI",
    "GLUE_QQP",
    "GLUE_RTE",
    "GLUE_WNLI",
    "SUPERGLUE_RTE",
    "DataPairCorpus",
    "DataPairDataset",
    "OpusParallelCorpus",
    "ParallelTextCorpus",
    "ParallelTextDataset",
    "UD_AFRIKAANS",
    "UD_ANCIENT_GREEK",
    "UD_ARABIC",
    "UD_ARMENIAN",
    "UD_BASQUE",
    "UD_BELARUSIAN",
    "UD_BULGARIAN",
    "UD_CATALAN",
    "UD_CHINESE",
    "UD_COPTIC",
    "UD_CROATIAN",
    "UD_CZECH",
    "UD_DANISH",
    "UD_DUTCH",
    "UD_ENGLISH",
    "UD_ESTONIAN",
    "UD_FAROESE",
    "UD_FINNISH",
    "UD_FRENCH",
    "UD_GALICIAN",
    "UD_GERMAN",
    "UD_GERMAN_HDT",
    "UD_GOTHIC",
    "UD_GREEK",
    "UD_HEBREW",
    "UD_HINDI",
    "UD_INDONESIAN",
    "UD_IRISH",
    "UD_ITALIAN",
    "UD_JAPANESE",
    "UD_KAZAKH",
    "UD_KOREAN",
    "UD_LATIN",
    "UD_LATVIAN",
    "UD_LITHUANIAN",
    "UD_LIVVI",
    "UD_MALTESE",
    "UD_MARATHI",
    "UD_NORTH_SAMI",
    "UD_NORWEGIAN",
    "UD_OLD_CHURCH_SLAVONIC",
    "UD_OLD_FRENCH",
    "UD_PERSIAN",
    "UD_POLISH",
    "UD_PORTUGUESE",
    "UD_ROMANIAN",
    "UD_RUSSIAN",
    "UD_SERBIAN",
    "UD_SLOVAK",
    "UD_SLOVENIAN",
    "UD_SPANISH",
    "UD_SWEDISH",
    "UD_TURKISH",
    "UD_WOLOF",
    "UniversalDependenciesCorpus",
    "UniversalDependenciesDataset",
]



import pytest

import flair
from flair.datasets import ClassificationCorpus
from flair.embeddings import DocumentPoolEmbeddings, FlairEmbeddings
from flair.models import TextClassifier
from flair.trainers import ModelTrainer


@pytest.mark.integration
def test_text_classifier_multi(results_base_path, tasks_base_path):
    flair.set_seed(123)

    flair_embeddings = FlairEmbeddings("news-forward-fast")

    corpus = ClassificationCorpus(
        tasks_base_path / "trivial" / "trivial_text_classification_single",
        label_type="city",
    )
    label_dict = corpus.make_label_dictionary(label_type="city")

    model: TextClassifier = TextClassifier(
        document_embeddings=DocumentPoolEmbeddings([flair_embeddings], fine_tune_mode="linear"),
        label_dictionary=label_dict,
        label_type="city",
    )

    trainer = ModelTrainer(model, corpus)
    trainer.train(results_base_path, mini_batch_size=2, max_epochs=1, shuffle=True)

    del model
    train_log_file = results_base_path / "training.log"
    assert train_log_file.exists()
    lines = train_log_file.read_text(encoding="utf-8").split("\n")
    expected_substrings = [
        "Device: ",
        "Corpus: ",
        "Parameters:",
        "- learning_rate: ",
        "- patience: ",
        "Embeddings storage mode:",
        "epoch 1 - iter",
        "EPOCH 1 done: loss",
        "Results:",
    ]
    for expected_substring in expected_substrings:
        assert any(expected_substring in line for line in lines), expected_substring



import logging
import re
import typing
from abc import ABC, abstractmethod
from collections import Counter, defaultdict
from functools import lru_cache
from operator import itemgetter
from pathlib import Path
from typing import Dict, List, Optional, Union, cast

import torch
from deprecated import deprecated
from torch.utils.data import Dataset
from torch.utils.data.dataset import ConcatDataset, Subset

import flair
from flair.file_utils import Tqdm

log = logging.getLogger("flair")


def _iter_dataset(dataset: Optional[Dataset]) -> typing.Iterable:
    if dataset is None:
        return []
    from flair.datasets import DataLoader

    return map(lambda x: x[0], DataLoader(dataset, batch_size=1, num_workers=0))


def _len_dataset(dataset: Optional[Dataset]) -> int:
    if dataset is None:
        return 0
    from flair.datasets import DataLoader

    loader = DataLoader(dataset, batch_size=1, num_workers=0)
    return len(loader)


class Dictionary:
    """
    This class holds a dictionary that maps strings to IDs, used to generate one-hot encodings of strings.
    """

    def __init__(self, add_unk=True):
        # init dictionaries
        self.item2idx: Dict[bytes, int] = {}
        self.idx2item: List[bytes] = []
        self.add_unk = add_unk
        self.multi_label = False
        self.span_labels = False
        # in order to deal with unknown tokens, add <unk>
        if add_unk:
            self.add_item("<unk>")

    def remove_item(self, item: str):

        bytes_item = item.encode("utf-8")
        if bytes_item in self.item2idx:
            self.idx2item.remove(bytes_item)
            del self.item2idx[bytes_item]

    def add_item(self, item: str) -> int:
        """
        add string - if already in dictionary returns its ID. if not in dictionary, it will get a new ID.
        :param item: a string for which to assign an id.
        :return: ID of string
        """
        bytes_item = item.encode("utf-8")
        if bytes_item not in self.item2idx:
            self.idx2item.append(bytes_item)
            self.item2idx[bytes_item] = len(self.idx2item) - 1
        return self.item2idx[bytes_item]

    def get_idx_for_item(self, item: str) -> int:
        """
        returns the ID of the string, otherwise 0
        :param item: string for which ID is requested
        :return: ID of string, otherwise 0
        """
        item_encoded = item.encode("utf-8")
        if item_encoded in self.item2idx.keys():
            return self.item2idx[item_encoded]
        elif self.add_unk:
            return 0
        else:
            log.error(f"The string '{item}' is not in dictionary! Dictionary contains only: {self.get_items()}")
            log.error(
                "You can create a Dictionary that handles unknown items with an <unk>-key by setting add_unk = True in the construction."
            )
            raise IndexError

    def get_idx_for_items(self, items: List[str]) -> List[int]:
        """
        returns the IDs for each item of the list of string, otherwise 0 if not found
        :param items: List of string for which IDs are requested
        :return: List of ID of strings
        """
        if not hasattr(self, "item2idx_not_encoded"):
            d = dict([(key.decode("UTF-8"), value) for key, value in self.item2idx.items()])
            self.item2idx_not_encoded = defaultdict(int, d)

        if not items:
            return []
        results = itemgetter(*items)(self.item2idx_not_encoded)
        if isinstance(results, int):
            return [results]
        return list(results)

    def get_items(self) -> List[str]:
        items = []
        for item in self.idx2item:
            items.append(item.decode("UTF-8"))
        return items

    def __len__(self) -> int:
        return len(self.idx2item)

    def get_item_for_index(self, idx):
        return self.idx2item[idx].decode("UTF-8")

    def set_start_stop_tags(self):
        self.add_item("<START>")
        self.add_item("<STOP>")

    def start_stop_tags_are_set(self):
        if {"<START>".encode(), "<STOP>".encode()}.issubset(self.item2idx.keys()):
            return True
        else:
            return False

    def save(self, savefile):
        import pickle

        with open(savefile, "wb") as f:
            mappings = {"idx2item": self.idx2item, "item2idx": self.item2idx}
            pickle.dump(mappings, f)

    def __setstate__(self, d):
        self.__dict__ = d
        # set 'add_unk' if the dictionary was created with a version of Flair older than 0.9
        if "add_unk" not in self.__dict__.keys():
            self.__dict__["add_unk"] = True if b"<unk>" in self.__dict__["idx2item"] else False

    @classmethod
    def load_from_file(cls, filename: Union[str, Path]):
        import pickle

        f = open(filename, "rb")
        mappings = pickle.load(f, encoding="latin1")
        idx2item = mappings["idx2item"]
        item2idx = mappings["item2idx"]
        f.close()

        # set 'add_unk' depending on whether <unk> is a key
        add_unk = True if b"<unk>" in idx2item else False

        dictionary: Dictionary = Dictionary(add_unk=add_unk)
        dictionary.item2idx = item2idx
        dictionary.idx2item = idx2item
        return dictionary

    @classmethod
    def load(cls, name: str):
        from flair.file_utils import cached_path

        hu_path: str = "https://flair.informatik.hu-berlin.de/resources/characters"
        if name == "chars" or name == "common-chars":
            char_dict = cached_path(f"{hu_path}/common_characters", cache_dir="datasets")
            return Dictionary.load_from_file(char_dict)

        if name == "chars-large" or name == "common-chars-large":
            char_dict = cached_path(f"{hu_path}/common_characters_large", cache_dir="datasets")
            return Dictionary.load_from_file(char_dict)

        if name == "chars-xl" or name == "common-chars-xl":
            char_dict = cached_path(f"{hu_path}/common_characters_xl", cache_dir="datasets")
            return Dictionary.load_from_file(char_dict)

        if name == "chars-lemmatizer" or name == "common-chars-lemmatizer":
            char_dict = cached_path(f"{hu_path}/common_characters_lemmatizer", cache_dir="datasets")
            return Dictionary.load_from_file(char_dict)

        return Dictionary.load_from_file(name)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, Dictionary):
            return False
        return self.item2idx == o.item2idx and self.idx2item == o.idx2item and self.add_unk == o.add_unk

    def __str__(self):
        tags = ", ".join(self.get_item_for_index(i) for i in range(min(len(self), 50)))
        return f"Dictionary with {len(self)} tags: {tags}"


class Label:
    """
    This class represents a label. Each label has a value and optionally a confidence score. The
    score needs to be between 0.0 and 1.0. Default value for the score is 1.0.
    """

    def __init__(self, data_point: "DataPoint", value: Optional[str], score: float = 1.0):
        self._value = value
        self._score = score
        self.data_point: DataPoint = data_point
        super().__init__()

    def set_value(self, value: str, score: float = 1.0):
        self.value = value
        self.score = score

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if not value and value != "":
            raise ValueError("Incorrect label value provided. Label value needs to be set.")
        else:
            self._value = value

    @property
    def score(self):
        return self._score

    @score.setter
    def score(self, score):
        self._score = score

    def to_dict(self):
        return {"value": self.value, "confidence": self.score}

    def __str__(self):
        return f"{self.data_point.unlabeled_identifier}{flair._arrow}{self._value} ({round(self._score, 4)})"

    @property
    def shortstring(self):
        return f'"{self.data_point.text}"/{self._value}'

    def __repr__(self):
        return f"'{self.data_point.unlabeled_identifier}'/'{self._value}' ({round(self._score, 4)})"

    def __eq__(self, other):
        return self.value == other.value and self.score == other.score and self.data_point == other.data_point

    def __hash__(self):
        return hash(self.__repr__())

    def __lt__(self, other):
        return self.data_point < other.data_point

    @property
    def labeled_identifier(self):
        return f"{self.data_point.unlabeled_identifier}/{self.value}"

    @property
    def unlabeled_identifier(self):
        return f"{self.data_point.unlabeled_identifier}"


class DataPoint:
    """
    This is the parent class of all data points in Flair (including Token, Sentence, Image, etc.). Each DataPoint
    must be embeddable (hence the abstract property embedding() and methods to() and clear_embeddings()). Also,
    each DataPoint may have Labels in several layers of annotation (hence the functions add_label(), get_labels()
    and the property 'label')
    """

    def __init__(self):
        self.annotation_layers = {}
        self._embeddings: Dict[str, torch.Tensor] = {}

    @property
    @abstractmethod
    def embedding(self):
        pass

    def set_embedding(self, name: str, vector: torch.Tensor):
        self._embeddings[name] = vector

    def get_embedding(self, names: Optional[List[str]] = None) -> torch.Tensor:
        embeddings = self.get_each_embedding(names)

        if embeddings:
            return torch.cat(embeddings, dim=0)

        return torch.tensor([], device=flair.device)

    def get_each_embedding(self, embedding_names: Optional[List[str]] = None) -> List[torch.Tensor]:
        embeddings = []
        for embed_name in sorted(self._embeddings.keys()):
            if embedding_names and embed_name not in embedding_names:
                continue
            embed = self._embeddings[embed_name].to(flair.device)
            embeddings.append(embed)
        return embeddings

    def to(self, device: str, pin_memory: bool = False):
        for name, vector in self._embeddings.items():
            if str(vector.device) != str(device):
                if pin_memory:
                    self._embeddings[name] = vector.to(device, non_blocking=True).pin_memory()
                else:
                    self._embeddings[name] = vector.to(device, non_blocking=True)

    def clear_embeddings(self, embedding_names: List[str] = None):
        if embedding_names is None:
            self._embeddings = {}
        else:
            for name in embedding_names:
                if name in self._embeddings.keys():
                    del self._embeddings[name]

    def has_label(self, type) -> bool:
        if type in self.annotation_layers:
            return True
        else:
            return False

    def add_label(self, typename: str, value: str, score: float = 1.0):

        if typename not in self.annotation_layers:
            self.annotation_layers[typename] = [Label(self, value, score)]
        else:
            self.annotation_layers[typename].append(Label(self, value, score))

        return self

    def set_label(self, typename: str, value: str, score: float = 1.0):
        self.annotation_layers[typename] = [Label(self, value, score)]
        return self

    def remove_labels(self, typename: str):
        if typename in self.annotation_layers.keys():
            del self.annotation_layers[typename]

    def get_label(self, label_type: str = None, zero_tag_value="O"):
        if len(self.get_labels(label_type)) == 0:
            return Label(self, zero_tag_value)
        return self.get_labels(label_type)[0]

    def get_labels(self, typename: str = None):
        if typename is None:
            return self.labels

        return self.annotation_layers[typename] if typename in self.annotation_layers else []

    @property
    def labels(self) -> List[Label]:
        all_labels = []
        for key in self.annotation_layers.keys():
            all_labels.extend(self.annotation_layers[key])
        return all_labels

    @property
    @abstractmethod
    def unlabeled_identifier(self):
        raise NotImplementedError

    def _printout_labels(self, main_label=None, add_score: bool = True):

        all_labels = []
        keys = [main_label] if main_label is not None else self.annotation_layers.keys()
        if add_score:
            for key in keys:
                all_labels.extend(
                    [
                        f"{label.value} ({round(label.score, 4)})"
                        for label in self.get_labels(key)
                        if label.data_point == self
                    ]
                )
            labels = "; ".join(all_labels)
            if labels != "":
                labels = flair._arrow + labels
        else:
            for key in keys:
                all_labels.extend([f"{label.value}" for label in self.get_labels(key) if label.data_point == self])
            labels = "/".join(all_labels)
            if labels != "":
                labels = "/" + labels
        return labels

    def __str__(self) -> str:
        return self.unlabeled_identifier + self._printout_labels()

    @property
    @abstractmethod
    def start_position(self) -> int:
        raise NotImplementedError

    @property
    @abstractmethod
    def end_position(self) -> int:
        raise NotImplementedError

    @property
    @abstractmethod
    def text(self):
        raise NotImplementedError

    @property
    def tag(self):
        return self.labels[0].value

    @property
    def score(self):
        return self.labels[0].score

    def __lt__(self, other):
        return self.start_position < other.start_position

    def __eq__(self, other):
        # TODO: does it make sense to exclude labels? Two data points of identical text (but different labels)
        #  would be equal now.
        return self.unlabeled_identifier == other.unlabeled_identifier

    def __hash__(self):
        return hash(self.unlabeled_identifier)


DT = typing.TypeVar("DT", bound=DataPoint)
DT2 = typing.TypeVar("DT2", bound=DataPoint)


class _PartOfSentence(DataPoint, ABC):
    def __init__(self, sentence):
        super().__init__()
        self.sentence: Sentence = sentence

    def _init_labels(self):
        if self.unlabeled_identifier in self.sentence._known_spans:
            self.annotation_layers = self.sentence._known_spans[self.unlabeled_identifier].annotation_layers
            del self.sentence._known_spans[self.unlabeled_identifier]
        self.sentence._known_spans[self.unlabeled_identifier] = self

    def add_label(self, typename: str, value: str, score: float = 1.0):
        super().add_label(typename, value, score)

        label = Label(self, value, score)
        if typename not in self.sentence.annotation_layers:
            self.sentence.annotation_layers[typename] = [label]
        else:
            self.sentence.annotation_layers[typename].append(label)

    def set_label(self, typename: str, value: str, score: float = 1.0):
        super().set_label(typename, value, score)
        self.sentence.annotation_layers[typename] = [Label(self, value, score)]
        return self

    def remove_labels(self, typename: str):

        # labels also need to be deleted at Sentence object
        for label in self.get_labels(typename):
            self.sentence.annotation_layers[typename].remove(label)

        # delete labels at object itself
        super(_PartOfSentence, self).remove_labels(typename)


class Token(_PartOfSentence):
    """
    This class represents one word in a tokenized sentence. Each token may have any number of tags. It may also point
    to its head in a dependency tree.
    """

    def __init__(
        self,
        text: str,
        head_id: int = None,
        whitespace_after: int = 1,
        start_position: int = 0,
        sentence=None,
    ):
        super().__init__(sentence=sentence)

        self.form: str = text
        self._internal_index: Optional[int] = None
        self.head_id: Optional[int] = head_id
        self.whitespace_after: int = whitespace_after

        self.start_pos = start_position
        self.end_pos = start_position + len(text)

        self._embeddings: Dict = {}
        self.tags_proba_dist: Dict[str, List[Label]] = {}

    @property
    def idx(self) -> int:
        if isinstance(self._internal_index, int):
            return self._internal_index
        else:
            raise ValueError

    @property
    def text(self):
        return self.form

    @property
    def unlabeled_identifier(self) -> str:
        return f'Token[{self.idx-1}]: "{self.text}"'

    def add_tags_proba_dist(self, tag_type: str, tags: List[Label]):
        self.tags_proba_dist[tag_type] = tags

    def get_tags_proba_dist(self, tag_type: str) -> List[Label]:
        if tag_type in self.tags_proba_dist:
            return self.tags_proba_dist[tag_type]
        return []

    def get_head(self):
        return self.sentence.get_token(self.head_id)

    @property
    def start_position(self) -> int:
        return self.start_pos

    @property
    def end_position(self) -> int:
        return self.end_pos

    @property
    def embedding(self):
        return self.get_embedding()

    def __repr__(self):
        return self.__str__()

    def add_label(self, typename: str, value: str, score: float = 1.0):
        """
        The Token is a special _PartOfSentence in that it may be initialized without a Sentence.
        Therefore, labels get added only to the Sentence if it exists
        """
        if self.sentence:
            super().add_label(typename=typename, value=value, score=score)
        else:
            DataPoint.add_label(self, typename=typename, value=value, score=score)

    def set_label(self, typename: str, value: str, score: float = 1.0):
        """
        The Token is a special _PartOfSentence in that it may be initialized without a Sentence.
        Therefore, labels get set only to the Sentence if it exists
        """
        if self.sentence:
            super().set_label(typename=typename, value=value, score=score)
        else:
            DataPoint.set_label(self, typename=typename, value=value, score=score)


class Span(_PartOfSentence):
    """
    This class represents one textual span consisting of Tokens.
    """

    def __init__(self, tokens: List[Token]):
        super().__init__(tokens[0].sentence)
        self.tokens = tokens
        super()._init_labels()

    @property
    def start_position(self) -> int:
        return self.tokens[0].start_position

    @property
    def end_position(self) -> int:
        return self.tokens[-1].end_position

    @property
    def text(self) -> str:
        return " ".join([t.text for t in self.tokens])

    @property
    def unlabeled_identifier(self) -> str:
        return f'Span[{self.tokens[0].idx -1}:{self.tokens[-1].idx}]: "{self.text}"'

    def __repr__(self):
        return self.__str__()

    def __getitem__(self, idx: int) -> Token:
        return self.tokens[idx]

    def __iter__(self):
        return iter(self.tokens)

    def __len__(self) -> int:
        return len(self.tokens)

    @property
    def embedding(self):
        pass


class Relation(_PartOfSentence):
    def __init__(self, first: Span, second: Span):
        super().__init__(sentence=first.sentence)
        self.first: Span = first
        self.second: Span = second
        super()._init_labels()

    def __repr__(self):
        return str(self)

    @property
    def tag(self):
        return self.labels[0].value

    @property
    def text(self):
        return f"{self.first.text} -> {self.second.text}"

    @property
    def unlabeled_identifier(self) -> str:
        return (
            f"Relation"
            f"[{self.first.tokens[0].idx-1}:{self.first.tokens[-1].idx}]"
            f"[{self.second.tokens[0].idx-1}:{self.second.tokens[-1].idx}]"
            f': "{self.text}"'
        )

    @property
    def start_position(self) -> int:
        return min(self.first.start_position, self.second.start_position)

    @property
    def end_position(self) -> int:
        return max(self.first.end_position, self.second.end_position)

    @property
    def embedding(self):
        pass


class Tokenizer(ABC):
    r"""An abstract class representing a :class:`Tokenizer`.

    Tokenizers are used to represent algorithms and models to split plain text into
    individual tokens / words. All subclasses should overwrite :meth:`tokenize`, which
    splits the given plain text into tokens. Moreover, subclasses may overwrite
    :meth:`name`, returning a unique identifier representing the tokenizer's
    configuration.
    """

    @abstractmethod
    def tokenize(self, text: str) -> List[str]:
        raise NotImplementedError()

    @property
    def name(self) -> str:
        return self.__class__.__name__


class Sentence(DataPoint):
    """
    A Sentence is a list of tokens and is used to represent a sentence or text fragment.
    """

    def __init__(
        self,
        text: Union[str, List[str]],
        use_tokenizer: Union[bool, Tokenizer] = True,
        language_code: str = None,
        start_position: int = 0,
    ):
        """
        Class to hold all meta related to a text (tokens, predictions, language code, ...)
        :param text: original string (sentence), or a list of string tokens (words)
        :param use_tokenizer: a custom tokenizer (default is :class:`SpaceTokenizer`)
            more advanced options are :class:`SegTokTokenizer` to use segtok or :class:`SpacyTokenizer`
            to use Spacy library if available). Check the implementations of abstract class Tokenizer or
            implement your own subclass (if you need it). If instead of providing a Tokenizer, this parameter
            is just set to True (deprecated), :class:`SegtokTokenizer` will be used.
        :param language_code: Language of the sentence
        :param start_position: Start char offset of the sentence in the superordinate document
        """
        super().__init__()

        self.tokens: List[Token] = []

        # private field for all known spans
        self._known_spans: Dict[str, _PartOfSentence] = {}

        self.language_code: Optional[str] = language_code

        self.start_pos = start_position
        self.end_pos = start_position + len(text)

        # the tokenizer used for this sentence
        if isinstance(use_tokenizer, Tokenizer):
            tokenizer = use_tokenizer

        elif type(use_tokenizer) == bool:
            from flair.tokenization import SegtokTokenizer, SpaceTokenizer

            tokenizer = SegtokTokenizer() if use_tokenizer else SpaceTokenizer()

        else:
            raise AssertionError("Unexpected type of parameter 'use_tokenizer'. Parameter should be bool or Tokenizer")

        # if text is passed, instantiate sentence with tokens (words)
        if isinstance(text, str):
            text = Sentence._handle_problem_characters(text)
            words = tokenizer.tokenize(text)
        else:
            words = text

        # determine token positions and whitespace_after flag
        current_offset = 0
        previous_word_offset = -1
        previous_token = None
        for word in words:
            try:
                word_offset = text.index(word, current_offset)
                start_position = word_offset
                delta_offset = start_position - current_offset
            except ValueError:
                word_offset = previous_word_offset + 1
                start_position = current_offset + 1 if current_offset > 0 else current_offset
                delta_offset = start_position - current_offset

            if word:
                token = Token(text=word, start_position=start_position)
                self.add_token(token)

            if previous_token is not None:
                previous_token.whitespace_after = delta_offset

            current_offset = word_offset + len(word)
            previous_word_offset = current_offset - 1
            previous_token = token

        # the last token has no whitespace after
        if len(self) > 0:
            self.tokens[-1].whitespace_after = 0

        # log a warning if the dataset is empty
        if text == "":
            log.warning("Warning: An empty Sentence was created! Are there empty strings in your dataset?")

        self.tokenized: Optional[str] = None

        # some sentences represent a document boundary (but most do not)
        self.is_document_boundary: bool = False

        # internal variables to denote position inside dataset
        self._previous_sentence: Optional[Sentence] = None
        self._next_sentence: Optional[Sentence] = None
        self._position_in_dataset: Optional[typing.Tuple[Dataset, int]] = None

    @property
    def unlabeled_identifier(self):
        return f'Sentence: "{self.to_tokenized_string()}"'

    def get_relations(self, type: str) -> List[Relation]:
        relations: List[Relation] = []
        for label in self.get_labels(type):
            if isinstance(label.data_point, Relation):
                relations.append(label.data_point)
        return relations

    def get_spans(self, type: str) -> List[Span]:
        spans: List[Span] = []
        for potential_span in self._known_spans.values():
            if isinstance(potential_span, Span) and potential_span.has_label(type):
                spans.append(potential_span)
        return sorted(spans)

    def get_token(self, token_id: int) -> Optional[Token]:
        for token in self.tokens:
            if token.idx == token_id:
                return token
        return None

    def add_token(self, token: Union[Token, str]):

        if isinstance(token, Token):
            assert token.sentence is None

        if type(token) is str:
            token = Token(token)
        token = cast(Token, token)

        # data with zero-width characters cannot be handled
        if token.text == "":
            return

        # set token idx and sentence
        token.sentence = self
        token._internal_index = len(self.tokens) + 1
        if token.start_position == 0 and len(self) > 0:
            token.start_pos = len(self.to_original_text()) + self[-1].whitespace_after
            token.end_pos = token.start_pos + len(token.text)

        # append token to sentence
        self.tokens.append(token)

        # register token annotations on sentence
        for typename in token.annotation_layers.keys():
            for label in token.get_labels(typename):
                if typename not in token.sentence.annotation_layers:
                    token.sentence.annotation_layers[typename] = [Label(token, label.value, label.score)]
                else:
                    token.sentence.annotation_layers[typename].append(Label(token, label.value, label.score))

    @property
    def embedding(self):
        return self.get_embedding()

    def to(self, device: str, pin_memory: bool = False):

        # move sentence embeddings to device
        super().to(device=device, pin_memory=pin_memory)

        # also move token embeddings to device
        for token in self:
            token.to(device, pin_memory)

    def clear_embeddings(self, embedding_names: List[str] = None):

        super().clear_embeddings(embedding_names)

        # clear token embeddings
        for token in self:
            token.clear_embeddings(embedding_names)

    @lru_cache(maxsize=1)  # cache last context, as training repeats calls
    def left_context(self, context_length: int, respect_document_boundaries: bool = True):
        sentence = self
        left_context: List[str] = []
        while True:
            sentence = sentence.previous_sentence()
            if sentence is None:
                break

            if respect_document_boundaries and sentence.is_document_boundary:
                break

            left_context = [t.text for t in sentence.tokens] + left_context
            if len(left_context) > context_length:
                left_context = left_context[-context_length:]
                break
        return left_context

    @lru_cache(maxsize=1)  # cache last context, as training repeats calls
    def right_context(self, context_length: int, respect_document_boundaries: bool = True):
        sentence = self
        right_context: List[str] = []
        while True:
            sentence = sentence.next_sentence()
            if sentence is None:
                break
            if respect_document_boundaries and sentence.is_document_boundary:
                break

            right_context += [t.text for t in sentence.tokens]
            if len(right_context) > context_length:
                right_context = right_context[:context_length]
                break
        return right_context

    def __str__(self):
        return self.to_tagged_string()

    def to_tagged_string(self, main_label=None) -> str:

        already_printed = [self]

        output = super().__str__()

        label_append = []
        for label in self.get_labels(main_label):
            if label.data_point in already_printed:
                continue
            label_append.append(
                f'"{label.data_point.text}"{label.data_point._printout_labels(main_label=main_label, add_score=False)}'
            )
            already_printed.append(label.data_point)

        if len(label_append) > 0:
            output += f"{flair._arrow}[" + ", ".join(label_append) + "]"

        return output

    @property
    def text(self):
        return self.to_tokenized_string()

    def to_tokenized_string(self) -> str:

        if self.tokenized is None:
            self.tokenized = " ".join([t.text for t in self.tokens])

        return self.tokenized

    def to_plain_string(self):
        plain = ""
        for token in self.tokens:
            plain += token.text
            if token.whitespace_after > 0:
                plain += token.whitespace_after * " "
        return plain.rstrip()

    def infer_space_after(self):
        """
        Heuristics in case you wish to infer whitespace_after values for tokenized text. This is useful for some old NLP
        tasks (such as CoNLL-03 and CoNLL-2000) that provide only tokenized data with no info of original whitespacing.
        :return:
        """
        last_token = None
        quote_count: int = 0
        # infer whitespace after field

        for token in self.tokens:
            if token.text == '"':
                quote_count += 1
                if quote_count % 2 != 0:
                    token.whitespace_after = 0
                elif last_token is not None:
                    last_token.whitespace_after = 0

            if last_token is not None:

                if token.text in [".", ":", ",", ";", ")", "n't", "!", "?"]:
                    last_token.whitespace_after = 0

                if token.text.startswith("'"):
                    last_token.whitespace_after = 0

            if token.text in ["("]:
                token.whitespace_after = 0

            last_token = token
        return self

    def to_original_text(self) -> str:
        str = ""
        pos = 0
        for t in self.tokens:
            while t.start_pos > pos:
                str += " "
                pos += 1

            str += t.text
            pos += len(t.text)

        return str

    def to_dict(self, tag_type: str = None):
        labels = []

        if tag_type:
            labels = [label.to_dict() for label in self.get_labels(tag_type)]
            return {"text": self.to_original_text(), tag_type: labels}

        if self.labels:
            labels = [label.to_dict() for label in self.labels]

        return {"text": self.to_original_text(), "all labels": labels}

    def get_span(self, start: int, stop: int):
        span_slice = slice(start, stop)
        return self[span_slice]

    @typing.overload
    def __getitem__(self, idx: int) -> Token:
        ...

    @typing.overload
    def __getitem__(self, s: slice) -> Span:
        ...

    def __getitem__(self, subscript):
        if isinstance(subscript, slice):
            return Span(self.tokens[subscript])
        else:
            return self.tokens[subscript]

    def __iter__(self):
        return iter(self.tokens)

    def __len__(self) -> int:
        return len(self.tokens)

    def __repr__(self):
        return self.__str__()

    def __copy__(self):
        s = Sentence()
        for token in self.tokens:
            nt = Token(token.text)
            for tag_type in token.tags:
                nt.add_label(
                    tag_type,
                    token.get_tag(tag_type).value,
                    token.get_tag(tag_type).score,
                )

            s.add_token(nt)
        return s

    @property
    def start_position(self) -> int:
        return 0

    @property
    def end_position(self) -> int:
        return len(self.to_original_text())

    def get_language_code(self) -> str:
        if self.language_code is None:
            import langdetect

            try:
                self.language_code = langdetect.detect(self.to_plain_string())
            except Exception:
                self.language_code = "en"

        return self.language_code

    @staticmethod
    def _handle_problem_characters(text: str) -> str:
        text = Sentence.__remove_zero_width_characters(text)
        text = Sentence.__restore_windows_1252_characters(text)
        return text

    @staticmethod
    def __remove_zero_width_characters(text: str) -> str:
        text = text.replace("\u200c", "")
        text = text.replace("\u200b", "")
        text = text.replace("\ufe0f", "")
        text = text.replace("\ufeff", "")
        return text

    @staticmethod
    def __restore_windows_1252_characters(text: str) -> str:
        def to_windows_1252(match):
            try:
                return bytes([ord(match.group(0))]).decode("windows-1252")
            except UnicodeDecodeError:
                # No character at the corresponding code point: remove it
                return ""

        return re.sub(r"[\u0080-\u0099]", to_windows_1252, text)

    def next_sentence(self):
        """
        Get the next sentence in the document (works only if context is set through dataloader or elsewhere)
        :return: next Sentence in document if set, otherwise None
        """
        if self._next_sentence is not None:
            return self._next_sentence

        if self._position_in_dataset is not None:
            dataset = self._position_in_dataset[0]
            index = self._position_in_dataset[1] + 1
            if index < len(dataset):
                return dataset[index]

        return None

    def previous_sentence(self):
        """
        Get the previous sentence in the document (works only if context is set through dataloader or elsewhere)
        :return: previous Sentence in document if set, otherwise None
        """
        if self._previous_sentence is not None:
            return self._previous_sentence

        if self._position_in_dataset is not None:
            dataset = self._position_in_dataset[0]
            index = self._position_in_dataset[1] - 1
            if index >= 0:
                return dataset[index]

        return None

    def is_context_set(self) -> bool:
        """
        Return True or False depending on whether context is set (for instance in dataloader or elsewhere)
        :return: True if context is set, else False
        """
        return self._previous_sentence is not None or self._position_in_dataset is not None

    def get_labels(self, label_type: str = None):

        # if no label if specified, return all labels
        if label_type is None:
            return sorted(self.labels)

        # if the label type exists in the Sentence, return it
        if label_type in self.annotation_layers:
            return sorted(self.annotation_layers[label_type])

        # return empty list if none of the above
        return []

    def remove_labels(self, typename: str):

        # labels also need to be deleted at all tokens
        for token in self:
            token.remove_labels(typename)

        # labels also need to be deleted at all known spans
        for span in self._known_spans.values():
            span.remove_labels(typename)

        # remove spans without labels
        self._known_spans = {k: v for k, v in self._known_spans.items() if len(v.labels) > 0}

        # delete labels at object itself
        super().remove_labels(typename)


class DataPair(DataPoint, typing.Generic[DT, DT2]):
    def __init__(self, first: DT, second: DT2):
        super().__init__()
        self.first = first
        self.second = second

    def to(self, device: str, pin_memory: bool = False):
        self.first.to(device, pin_memory)
        self.second.to(device, pin_memory)

    def clear_embeddings(self, embedding_names: List[str] = None):
        self.first.clear_embeddings(embedding_names)
        self.second.clear_embeddings(embedding_names)

    @property
    def embedding(self):
        return torch.cat([self.first.embedding, self.second.embedding])

    def __len__(self):
        return len(self.first) + len(self.second)

    @property
    def unlabeled_identifier(self):
        return f"DataPair: '{self.first.unlabeled_identifier}' + '{self.second.unlabeled_identifier}'"

    @property
    def start_position(self) -> int:
        return self.first.start_position

    @property
    def end_position(self) -> int:
        return self.first.end_position

    @property
    def text(self):
        return self.first.text + " || " + self.second.text


TextPair = DataPair[Sentence, Sentence]


class Image(DataPoint):
    def __init__(self, data=None, imageURL=None):
        super().__init__()

        self.data = data
        self._embeddings: Dict = {}
        self.imageURL = imageURL

    @property
    def embedding(self):
        return self.get_embedding()

    def __str__(self):
        image_repr = self.data.size() if self.data else ""
        image_url = self.imageURL if self.imageURL else ""

        return f"Image: {image_repr} {image_url}"

    @property
    def start_position(self) -> int:
        pass

    @property
    def end_position(self) -> int:
        pass

    @property
    def text(self):
        pass

    @property
    def unlabeled_identifier(self):
        pass


class FlairDataset(Dataset):
    @abstractmethod
    def is_in_memory(self) -> bool:
        pass


class Corpus:
    def __init__(
        self,
        train: Dataset = None,
        dev: Dataset = None,
        test: Dataset = None,
        name: str = "corpus",
        sample_missing_splits: Union[bool, str] = True,
    ):
        # set name
        self.name: str = name

        # abort if no data is provided
        if not train and not dev and not test:
            raise RuntimeError("No data provided when initializing corpus object.")

        # sample test data from train if none is provided
        if test is None and sample_missing_splits and train and not sample_missing_splits == "only_dev":
            train_length = _len_dataset(train)
            test_size: int = round(train_length / 10)
            test, train = randomly_split_into_two_datasets(train, test_size)

        # sample dev data from train if none is provided
        if dev is None and sample_missing_splits and train and not sample_missing_splits == "only_test":
            train_length = _len_dataset(train)
            dev_size: int = round(train_length / 10)
            dev, train = randomly_split_into_two_datasets(train, dev_size)

        # set train dev and test data
        self._train: Optional[Dataset] = train
        self._test: Optional[Dataset] = test
        self._dev: Optional[Dataset] = dev

    @property
    def train(self) -> Optional[Dataset]:
        return self._train

    @property
    def dev(self) -> Optional[Dataset]:
        return self._dev

    @property
    def test(self) -> Optional[Dataset]:
        return self._test

    def downsample(
        self,
        percentage: float = 0.1,
        downsample_train=True,
        downsample_dev=True,
        downsample_test=True,
    ):

        if downsample_train and self._train is not None:
            self._train = self._downsample_to_proportion(self._train, percentage)

        if downsample_dev and self._dev is not None:
            self._dev = self._downsample_to_proportion(self._dev, percentage)

        if downsample_test and self._test is not None:
            self._test = self._downsample_to_proportion(self._test, percentage)

        return self

    def filter_empty_sentences(self):
        log.info("Filtering empty sentences")
        if self._train is not None:
            self._train = Corpus._filter_empty_sentences(self._train)
        if self._test is not None:
            self._test = Corpus._filter_empty_sentences(self._test)
        if self._dev is not None:
            self._dev = Corpus._filter_empty_sentences(self._dev)
        log.info(self)

    def filter_long_sentences(self, max_charlength: int):
        log.info("Filtering long sentences")
        if self._train is not None:
            self._train = Corpus._filter_long_sentences(self._train, max_charlength)
        if self._test is not None:
            self._test = Corpus._filter_long_sentences(self._test, max_charlength)
        if self._dev is not None:
            self._dev = Corpus._filter_long_sentences(self._dev, max_charlength)
        log.info(self)

    @staticmethod
    def _filter_long_sentences(dataset, max_charlength: int) -> Dataset:

        # find out empty sentence indices
        empty_sentence_indices = []
        non_empty_sentence_indices = []

        for index, sentence in Tqdm.tqdm(enumerate(_iter_dataset(dataset))):
            if len(sentence.to_plain_string()) > max_charlength:
                empty_sentence_indices.append(index)
            else:
                non_empty_sentence_indices.append(index)

        # create subset of non-empty sentence indices
        subset = Subset(dataset, non_empty_sentence_indices)

        return subset

    @staticmethod
    def _filter_empty_sentences(dataset) -> Dataset:

        # find out empty sentence indices
        empty_sentence_indices = []
        non_empty_sentence_indices = []

        for index, sentence in enumerate(_iter_dataset(dataset)):
            if len(sentence) == 0:
                empty_sentence_indices.append(index)
            else:
                non_empty_sentence_indices.append(index)

        # create subset of non-empty sentence indices
        subset = Subset(dataset, non_empty_sentence_indices)

        return subset

    def make_vocab_dictionary(self, max_tokens=-1, min_freq=1) -> Dictionary:
        """
        Creates a dictionary of all tokens contained in the corpus.
        By defining `max_tokens` you can set the maximum number of tokens that should be contained in the dictionary.
        If there are more than `max_tokens` tokens in the corpus, the most frequent tokens are added first.
        If `min_freq` is set the a value greater than 1 only tokens occurring more than `min_freq` times are considered
        to be added to the dictionary.
        :param max_tokens: the maximum number of tokens that should be added to the dictionary (-1 = take all tokens)
        :param min_freq: a token needs to occur at least `min_freq` times to be added to the dictionary (-1 = there is no limitation)
        :return: dictionary of tokens
        """
        tokens = self._get_most_common_tokens(max_tokens, min_freq)

        vocab_dictionary: Dictionary = Dictionary()
        for token in tokens:
            vocab_dictionary.add_item(token)

        return vocab_dictionary

    def _get_most_common_tokens(self, max_tokens, min_freq) -> List[str]:
        tokens_and_frequencies = Counter(self._get_all_tokens())

        tokens: List[str] = []
        for token, freq in tokens_and_frequencies.most_common():
            if (min_freq != -1 and freq < min_freq) or (max_tokens != -1 and len(tokens) == max_tokens):
                break
            tokens.append(token)
        return tokens

    def _get_all_tokens(self) -> List[str]:
        assert self.train
        tokens = list(map((lambda s: s.tokens), _iter_dataset(self.train)))
        tokens = [token for sublist in tokens for token in sublist]
        return list(map((lambda t: t.text), tokens))

    @staticmethod
    def _downsample_to_proportion(dataset: Dataset, proportion: float):

        sampled_size: int = round(_len_dataset(dataset) * proportion)
        splits = randomly_split_into_two_datasets(dataset, sampled_size)
        return splits[0]

    def obtain_statistics(self, label_type: str = None, pretty_print: bool = True) -> Union[dict, str]:
        """
        Print statistics about the class distribution (only labels of sentences are taken into account) and sentence
        sizes.
        """
        json_data = {
            "TRAIN": self._obtain_statistics_for(self.train, "TRAIN", label_type),
            "TEST": self._obtain_statistics_for(self.test, "TEST", label_type),
            "DEV": self._obtain_statistics_for(self.dev, "DEV", label_type),
        }
        if pretty_print:
            import json

            return json.dumps(json_data, indent=4)
        return json_data

    @staticmethod
    def _obtain_statistics_for(sentences, name, tag_type) -> dict:
        if len(sentences) == 0:
            return {}

        classes_to_count = Corpus._count_sentence_labels(sentences)
        tags_to_count = Corpus._count_token_labels(sentences, tag_type)
        tokens_per_sentence = Corpus._get_tokens_per_sentence(sentences)

        label_size_dict = {}
        for label, c in classes_to_count.items():
            label_size_dict[label] = c

        tag_size_dict = {}
        for tag, c in tags_to_count.items():
            tag_size_dict[tag] = c

        return {
            "dataset": name,
            "total_number_of_documents": len(sentences),
            "number_of_documents_per_class": label_size_dict,
            "number_of_tokens_per_tag": tag_size_dict,
            "number_of_tokens": {
                "total": sum(tokens_per_sentence),
                "min": min(tokens_per_sentence),
                "max": max(tokens_per_sentence),
                "avg": sum(tokens_per_sentence) / len(sentences),
            },
        }

    @staticmethod
    def _get_tokens_per_sentence(sentences):
        return list(map(lambda x: len(x.tokens), sentences))

    @staticmethod
    def _count_sentence_labels(sentences):
        label_count = defaultdict(lambda: 0)
        for sent in sentences:
            for label in sent.labels:
                label_count[label.value] += 1
        return label_count

    @staticmethod
    def _count_token_labels(sentences, label_type):
        label_count = defaultdict(lambda: 0)
        for sent in sentences:
            for token in sent.tokens:
                if label_type in token.annotation_layers.keys():
                    label = token.get_tag(label_type)
                    label_count[label.value] += 1
        return label_count

    def __str__(self) -> str:
        return "Corpus: %d train + %d dev + %d test sentences" % (
            _len_dataset(self.train) if self.train else 0,
            _len_dataset(self.dev) if self.dev else 0,
            _len_dataset(self.test) if self.test else 0,
        )

    def make_label_dictionary(self, label_type: str, min_count: int = -1) -> Dictionary:
        """
        Creates a dictionary of all labels assigned to the sentences in the corpus.
        :return: dictionary of labels
        """
        label_dictionary: Dictionary = Dictionary(add_unk=True)
        label_dictionary.span_labels = False

        assert self.train
        datasets = [self.train]

        data: ConcatDataset = ConcatDataset(datasets)

        log.info("Computing label dictionary. Progress:")

        sentence_label_type_counter: typing.Counter[str] = Counter()
        label_value_counter: typing.Counter[str] = Counter()
        all_sentence_labels: List[str] = []
        for sentence in Tqdm.tqdm(_iter_dataset(data)):

            # count all label types per sentence
            sentence_label_type_counter.update(sentence.annotation_layers.keys())

            # go through all labels of label_type and count values
            labels = sentence.get_labels(label_type)
            for label in labels:
                if label.value not in all_sentence_labels:
                    label_value_counter[label.value] += 1

                # check if there are any span labels
                if type(label.data_point) == Span and len(label.data_point) > 1:
                    label_dictionary.span_labels = True

            if not label_dictionary.multi_label:
                if len(labels) > 1:
                    label_dictionary.multi_label = True

        # if an unk threshold is set, UNK all label values below this threshold
        erfasst_count = 0
        unked_count = 0
        for label, count in label_value_counter.most_common():
            if count >= min_count:
                label_dictionary.add_item(label)
                erfasst_count += count
            else:
                unked_count += count

        if len(label_dictionary.idx2item) == 0 or (
            len(label_dictionary.idx2item) == 1 and "<unk>" in label_dictionary.get_items()
        ):
            log.error(f"ERROR: You specified label_type='{label_type}' which is not in this dataset!")
            contained_labels = ", ".join(
                [f"'{label[0]}' (in {label[1]} sentences)" for label in sentence_label_type_counter.most_common()]
            )
            log.error(f"ERROR: The corpus contains the following label types: {contained_labels}")
            raise Exception

        log.info(
            f"Dictionary created for label '{label_type}' with {len(label_dictionary)} "
            f"values: {', '.join([label[0] + f' (seen {label[1]} times)' for label in label_value_counter.most_common(20)])}"
        )

        if unked_count > 0:
            log.info(f" - at UNK threshold {min_count}, {unked_count} instances are UNK'ed and {erfasst_count} remain")

        return label_dictionary

    def get_label_distribution(self):
        class_to_count = defaultdict(lambda: 0)
        for sent in self.train:
            for label in sent.labels:
                class_to_count[label.value] += 1
        return class_to_count

    def get_all_sentences(self) -> ConcatDataset:
        parts = []
        if self.train:
            parts.append(self.train)
        if self.dev:
            parts.append(self.dev)
        if self.test:
            parts.append(self.test)
        return ConcatDataset(parts)

    @deprecated(version="0.8", reason="Use 'make_label_dictionary' instead.")
    def make_tag_dictionary(self, tag_type: str) -> Dictionary:

        # Make the tag dictionary
        tag_dictionary: Dictionary = Dictionary(add_unk=False)
        tag_dictionary.add_item("O")
        for sentence in _iter_dataset(self.get_all_sentences()):
            for token in sentence.tokens:
                tag_dictionary.add_item(token.get_tag(tag_type).value)
        tag_dictionary.add_item("<START>")
        tag_dictionary.add_item("<STOP>")
        return tag_dictionary


class MultiCorpus(Corpus):
    def __init__(self, corpora: List[Corpus], name: str = "multicorpus", **corpusargs):
        self.corpora: List[Corpus] = corpora

        train_parts = []
        dev_parts = []
        test_parts = []
        for corpus in self.corpora:
            if corpus.train:
                train_parts.append(corpus.train)
            if corpus.dev:
                dev_parts.append(corpus.dev)
            if corpus.test:
                test_parts.append(corpus.test)

        super(MultiCorpus, self).__init__(
            ConcatDataset(train_parts) if len(train_parts) > 0 else None,
            ConcatDataset(dev_parts) if len(dev_parts) > 0 else None,
            ConcatDataset(test_parts) if len(test_parts) > 0 else None,
            name=name,
            **corpusargs,
        )

    def __str__(self):
        output = (
            f"MultiCorpus: "
            f"{len(self.train) if self.train else 0} train + "
            f"{len(self.dev) if self.dev else 0} dev + "
            f"{len(self.test) if self.test else 0} test sentences\n - "
        )
        output += "\n - ".join([f"{type(corpus).__name__} {str(corpus)} - {corpus.name}" for corpus in self.corpora])
        return output


def iob2(tags):
    """
    Check that tags have a valid IOB format.
    Tags in IOB1 format are converted to IOB2.
    """
    for i, tag in enumerate(tags):
        if tag.value == "O":
            continue
        split = tag.value.split("-")
        if len(split) != 2 or split[0] not in ["I", "B"]:
            return False
        if split[0] == "B":
            continue
        elif i == 0 or tags[i - 1].value == "O":  # conversion IOB1 to IOB2
            tags[i].value = "B" + tag.value[1:]
        elif tags[i - 1].value[1:] == tag.value[1:]:
            continue
        else:  # conversion IOB1 to IOB2
            tags[i].value = "B" + tag.value[1:]
    return True


def iob_iobes(tags):
    """
    IOB -> IOBES
    """
    for i, tag in enumerate(tags):
        if tag.value == "O" or tag.value == "":
            tag.value = "O"
            continue
        t, label = tag.value.split("-", 1)
        if len(tags) == i + 1 or tags[i + 1].value == "O":
            next_same = False
        else:
            nt, next_label = tags[i + 1].value.split("-", 1)
            next_same = nt == "I" and next_label == label
        if t == "B":
            if not next_same:
                tag.value = tag.value.replace("B-", "S-")
        elif t == "I":
            if not next_same:
                tag.value = tag.value.replace("I-", "E-")
        else:
            raise Exception("Invalid IOB format!")


def randomly_split_into_two_datasets(dataset, length_of_first):
    import random

    indices = [i for i in range(len(dataset))]
    random.shuffle(indices)

    first_dataset = indices[:length_of_first]
    second_dataset = indices[length_of_first:]
    first_dataset.sort()
    second_dataset.sort()

    return Subset(dataset, first_dataset), Subset(dataset, second_dataset)
